import numpy as np
from sklearn.preprocessing import MinMaxScaler
from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix


# def _score_offline_2d_average(self, x_scaled: list) -> list:
#     """Core implementation of offline score calculation with average features. FOR 2D or more D.
#
#     Args:
#         x_scaled: normalized time series if is_scaled False
#         self.dynamic_mode: type of model to check differences in the sequence
#
#     Returns:
#         filtered_score: represented a list of values with 0 and 1 where 1 is an anomaly if view is True
#
#     """
#     score_list = []
#     if not self.dynamic_mode:
#         step = self.lag
#         start_idx = step
#         end_idx = len(x_scaled[0]) - step
#         horm_hist = None
#         average_features = self.average_features_2d(end_idx, start_idx, x_scaled)
#
#         for current_index in range(start_idx, end_idx, step):
#             current_features = self.current_features_2d(current_index, step, x_scaled)
#             current_features = np.reshape(current_features, (len(x_scaled), 7))
#             if horm_hist is None:
#                 horm_hist = np.linalg.norm(distance_matrix(average_features.T, current_features.T), 2)
#             score_list.append(np.linalg.norm(distance_matrix(average_features.T, current_features.T), 2))
#     else:
#         start_idx = self.ts_window_length + self.lag
#         end_idx = len(x_scaled[0]) - self.ts_window_length - 1
#         horm_hist = None
#         x_history_arr = []
#         average_features = self.average_features_2d(end_idx, start_idx, x_scaled)
#         x_history = None
#         for ts_number in range(len(average_features)):
#             x_history = self.spectrum_extractor.ts_vector_to_trajectory_matrix(
#                 timeseries=average_features[ts_number],
#                 K=self.ts_window_length - self.L + 1,
#                 L=len(average_features[ts_number]))
#             x_history_arr.extend(x_history)
#
#         for t in range(start_idx, end_idx, self.lag):  # get Hankel matrix
#             if horm_hist is None:
#                 horm_hist = np.linalg.norm(x_history, 1)
#
#             current_features = self.current_features_2d(t, self.ts_window_length, x_scaled)
#             current_features = current_features.reshape(current_features.shape[0],
#                                                         (current_features.shape[1] * current_features.shape[2]))
#
#             x_test_arr = []
#             x_test = None
#             for ts_number in range(len(current_features)):
#                 x_test = self.spectrum_extractor.ts_vector_to_trajectory_matrix(
#                     timeseries=current_features[ts_number],
#                     K=self.ts_window_length - self.L + 1,
#                     L=len(current_features[ts_number]))
#                 x_test_arr.extend(x_test)
#
#             if self.n_components is None:
#                 self._n_components(x_test, x_history)
#
#             score_list.append(
#                 self._sst_svd(x_test_arr, x_history_arr))
#     score_diff = np.diff(score_list)
#     q_95 = np.quantile(score_diff, self.quantile_rate)
#     filtered_score = score_diff
#     self.n_components = None
#     if self.view:
#         filtered_score = list(map(lambda _x: 1 if _x > q_95 else 0, score_diff))
#     return filtered_score
#
#
# def _score_offline_2d(self, x_scaled: list = None) -> list:
#     """Core implementation of offline score calculation. FOR 2D or more D.
#
#     Args:
#         x_scaled: normalized time series if is_scaled False
#         self.dynamic_mode: type of model to check differences in the sequence
#
#     Returns:
#         filtered_score: represented a list of values with 0 and 1 where 1 is an anomaly if view is True
#
#     """
#     norm_list_real = []
#     horm_hist = None
#     if self.dynamic_mode:
#         step = 1 * self.ts_window_length
#         start_idx = step
#         end_idx = len(x_scaled[0]) - step
#
#         current_index = start_idx
#         first_window = self._get_window_from_ts_complex(x_scaled, current_index, current_index + step)
#         first_features = self._get_features_vector_from_window(first_window)
#         first_features = np.asarray(first_features)
#         first_features = np.reshape(first_features, (len(x_scaled), 7))
#
#         for current_index in range(start_idx, end_idx, step):
#             current_window = self._get_window_from_ts_complex(x_scaled, current_index, current_index + step)
#             current_features = self._get_features_vector_from_window(current_window)
#             current_features = np.asarray(current_features)
#             current_features = np.reshape(current_features, (len(x_scaled), 7))
#             if horm_hist is None:
#                 horm_hist = np.linalg.norm(distance_matrix(first_features.T, current_features.T), 2)
#             norm_list_real.append(np.linalg.norm(distance_matrix(first_features.T, current_features.T), 2))
#     else:
#         raise ValueError("Function dose not work when dynamic == False (FOR 2D or more D)")
#
#     score_list = [horm_hist] + norm_list_real
#     score_diff = np.diff(score_list)
#     return score_diff

class SingularSpectrumTransformation:
    """SingularSpectrumTransformation class.

    Change point detection with Singular Spectrum Transformation.

    Note:
        In case of 1D time series to find appropriate hyperparameters value
         for ts_window_length and trajectory_window_length
         are recommended to use WSSAlgorithms.py, WindowSizeSelection class.

    Parameters:
        model_hyperparams:
                        n_components: Number of principal components to keep from
                            functional principal component analysis. Defaults to 3.
                        window_length: Regularization object to be applied.
                        trajectory_window_length: .
    Attributes:

    """

    def __init__(self, model_hyperparams: dict = None):

        if model_hyperparams is not None:
            self.n_components = model_hyperparams['n_components']
            self.window_length = model_hyperparams['window_length']
            self.trajectory_window_length = model_hyperparams['trajectory_window_length']
            self.dynamic_mode = model_hyperparams['dynamic_mode']
            self.lag = model_hyperparams['delay_lag']

        if self.lag is None:
            self.lag = np.round(self.window_length / 2)
        if self.n_components is None:
            self.n_components = 2

    def _scale_ts(self, time_series: np.ndarray):
        time_series_scaled = MinMaxScaler(feature_range=(1, 2)) \
                                 .fit_transform(time_series.reshape(-1, 1))[:, 0]
        return time_series_scaled

    def fit(self, train_features: np.ndarray) -> list:
        if len(train_features) > 1:
            list_of_history, list_of_current = [], []
            for time_series in train_features:
                history_state, current_state = self._fit(train_features=time_series)
                list_of_history.append(history_state)
                list_of_current.append(current_state)
            return list_of_history, list_of_current
        else:
            return self._fit(train_features=train_features)

    def _fit(self, train_features: np.ndarray) -> list:
        """Core implementation of offline score calculation.

        Args:
            self.dynamic_mode: mode for SST metrics calculation.
            train_features: input 1D time series data.

        Returns
            score: 1d array change point score with 1 and 0 if view True.

        """
        train_features = np.array(train_features)
        start_idx, end_idx = self.window_length + self.lag + 1, train_features.size + 1
        test_features = []
        if self.dynamic_mode:
            self.model = []
            for t in range(start_idx, end_idx):
                start_idx_hist, end_idx_hist = t - self.window_length - self.lag, t - self.lag
                start_idx_test, end_idx_test = t - self.window_length, t
                x_history = HankelMatrix(time_series=train_features[start_idx_hist:end_idx_hist],
                                         window_length=self.trajectory_window_length)
                x_test = HankelMatrix(time_series=train_features[start_idx_test:end_idx_test],
                                      window_length=self.trajectory_window_length)
                self.model.append(x_history)
                test_features.append(x_test)

        else:
            self.model = HankelMatrix(time_series=train_features[:start_idx],
                                      window_length=self.trajectory_window_length)
            for t in range(start_idx, end_idx):
                start_idx_test, end_idx_test = t - self.lag, t
                x_test = HankelMatrix(time_series=train_features[start_idx_test:end_idx_test],
                                      window_length=self.trajectory_window_length)
                test_features.append(x_test)

        return self.model, test_features

    def predict(self, test_features: list, target: list) -> list:
        if len(test_features) > 1:
            residual_list = []
            for history_data, target_data in zip(test_features, target):
                residual = self._predict(history_data, target_data)
                residual_list.append(residual)
            return residual_list
        else:
            return self._predict(test_features=test_features,target=target)

    def _predict(self, test_features: HankelMatrix, target: HankelMatrix) -> list:
        """Core implementation of offline score calculation.

        Args:
            self.dynamic_mode: mode for SST metrics calculation.
            _x: input 1D time series data.

        Returns
            score: 1d array change point score with 1 and 0 if view True.
            :param target:
            :param test_features:

        """

        if self.dynamic_mode:
            dynamic_sst = lambda trajectory_data_tuple: self._sst_svd(trajectory_data_tuple[0].trajectory_matrix,
                                                                      trajectory_data_tuple[1].trajectory_matrix)
            trajectory_data_tuple = list(zip(test_features, target))
            score_list = list(map(dynamic_sst, trajectory_data_tuple))
        else:
            static_sst = lambda x_history: self._sst_svd(self.model.trajectory_matrix, x_history.trajectory_matrix)
            score_list = list(map(static_sst, test_features))
        residual = np.diff(score_list)
        return residual

    def _sst_svd(self, x_test: list = None, x_history: list = None) -> float:
        """Singular value decomposition to count distance score between matrix

        Args:
            x_test: current matrix of features
            x_history: historical matrix of features

        Returns
            1 - s[0]: distance score between two matrix

        """
        u_test, s_test, _ = np.linalg.svd(x_test, full_matrices=False)
        u_history, s_hist, _ = np.linalg.svd(x_history, full_matrices=False)
        s_cov = u_test[:, :self.n_components].T @ u_history[:, :self.n_components]
        u_cov, s, _ = np.linalg.svd(s_cov, full_matrices=False)
        return 1 - s[0]



from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader


class ModelTestingModule:
    def __init__(self, model):
        self.model = model

    def extract_from_binary(self, dataset_name):
        train, test = DataLoader(dataset_name).load_data()
        self.test_target = test[1]
        self.train_target = train[1]
        train_feats = self.model.get_features(train[0], dataset_name)
        test_feats = self.model.get_features(test[0], dataset_name)
        return train_feats, test_feats

    def extract_from_multi_class(self, dataset_name):
        train, test = DataLoader(dataset_name).load_data()
        self.test_target = test[1]
        self.train_target = train[1]
        train_feats = self.model.get_features(train[0], dataset_name)
        test_feats = self.model.get_features(test[0], dataset_name)
        return train_feats, test_feats

    def fit(self, timeout: int):
        self.model.fit(timeout)

    def visualise(self, vis_data):
        self.model.visualise(vis_data)



import numpy as np


def quantile_filter(input_data, predicted_data, threshold: float = 0.9, lp_norm: int = 1):
    reconstruction_error = np.linalg.norm(input_data - predicted_data, lp_norm, axis=1) / np.linalg.norm(
        input_data, lp_norm, axis=1)
    quantile = np.quantile(reconstruction_error, threshold)
    outlier_idx = [np.where(np.isclose(reconstruction_error, idx_outlier))[0][0]
                   for idx_outlier in reconstruction_error[reconstruction_error > quantile]]
    return outlier_idx



import numpy as np


def quantile(column, q: str):
    return np.quantile(a=column, q=q)


def softmax(w, theta=1.0) -> np.ndarray:
    """Takes a vector w of S N-element and returns a vectors where each column
        of the vector sums to 1, with elements exponentially proportional to the
        respective elements in N.

        Args:
            w: array of shape = [N,  M].
            theta: float, parameter, used as a multiplier prior to exponentiation (default = 1.0).

        Returns:
            array of shape = [N, M]. Which the sum of each row sums to 1 and the elements are exponentially
            proportional to the respective elements in N

        """
    w = np.atleast_2d(w)
    e = np.exp(np.array(w) / theta)
    dist = e / np.sum(e, axis=1).reshape(-1, 1)
    return dist


stat_methods_default = {
    'mean_': np.mean,
    'median_': np.median,
    'std_': np.std,
    'var_': np.var,
    'q5_': quantile,
    'q25_': quantile,
    'q75_': quantile,
    'q95_': quantile,
}

stat_methods_ensemble = {
    'MeanEnsemble': np.mean,
    'MedianEnsemble': np.median,
    'MinEnsemble': np.min,
    'MaxEnsemble': np.max,
    'ProductEnsemble': np.prod
}

stat_methods_full = {
    'mean_': np.mean,
    'median_': np.median,
    'lambda_less_zero': lambda x: x < 0.01,
    'std_': np.std,
    'var_': np.var,
    'max': np.max,
    'min': np.min,
    'q5_': quantile,
    'q25_': quantile,
    'q75_': quantile,
    'q95_': quantile,
    'sum_': np.sum,
    'dif_': np.diff
}

hyper_param_dict = {'statistical_methods': stat_methods_default,
                    'statistical_methods_extra': stat_methods_full,
                    'stat_methods_ensemble': stat_methods_ensemble}


def select_hyper_param(param_name):
    return hyper_param_dict[param_name]



from typing import Mapping

from scipy import spatial

from fedot_ind.core.models.detection.abstract_objects.AbstractDataOperation import AbstractDataOperation
from fedot_ind.core.models.detection.abstract_objects.FileObject import FileObject
from fedot_ind.core.models.detection.utils.get_time import time_now
from fedot_ind.core.models.detection.utils.math_utils import NormalizeData

"""

input format:

    dict with "data" and "lables" fields

Output 
    the same dict but with additional list of window
    
"""
from typing import List


class Window:
    def __init__(self,
                 start: int,
                 end: int,
                 data_ts: Mapping[str, List[float]]
                 ) -> None:
        self.start: int = start
        self.end: int = end
        self.data_ts: Mapping[str, List[float]] = data_ts
        self.vector_result: float = 0

    def get_vector_predict(self) -> None:
        point_array: List[List[float]] = []
        keys = list(self.data_ts.keys())
        for i in range(0, self.end - self.start):
            point = []
            for key in keys:
                point.append(self.data_ts[key][i])
            point_array.append(point)

        cosine_array = []
        # last_point = point_array[-1]
        first_point = point_array[0]
        for i in range(0, len(point_array)):
            current_point = point_array[i]
            res = spatial.distance.cosine(current_point, first_point)
            cosine_array.append(res)
        avg = sum(cosine_array) / len(cosine_array)
        self.vector_result = avg  # ** 2


class AngleBasedDetector(AbstractDataOperation):

    def __init__(self, window_len: int = 100,
                 step: int = None,
                 detector_name: str = "Angle Based Detector"):
        self.window_len: list = window_len
        self.name: str = detector_name
        self.print_logs: bool = True
        if step is None:
            self.step: int = int(window_len / 4)
        else:
            self.step: int = step
        self.windows: List[Window] = []

    def set_settings(self):
        self.output_predicts: list = []
        self._print_logs("Settings was set.")

    def load_data(self, data_object) -> None:
        self._print_logs("Data read!")
        self.data_object: FileObject = data_object

    def run_operation(self) -> None:
        self._print_logs("Start detection...")
        self._run_detector()
        self._print_logs("Detection finished!")

    def return_new_data(self):
        return self.data_object

    def _split_data_to_windows(self,
                               ts: Mapping[str, List[float]],
                               len_of_data: int,
                               keys: List[str]
                               ) -> List[Window]:
        """
        This method split time series into windows

        Args:
            ts (Mapping[str, List[float]]): data
            len_of_data (int): len of data in <ts>
            keys (List[str]): keys of <ts> dict

        Returns:
            List[Window]: windows
        """
        start_idx: int = 0
        end_idx: int = len_of_data - self.window_len
        temp_window_list: List[Window] = []
        for i in range(start_idx, end_idx, self.step):
            temp_window_data: Mapping[str, List[float]] = {}
            for key in keys:
                temp_window_data[key] = []
                for j in range(i, i + self.window_len):
                    temp_window_data[key].append(ts[key][j])
            temp_window_list.append(Window(i, i + self.window_len, temp_window_data))
        return temp_window_list

    def _turn_windows_to_ts(self, windows: List[Window], demanded_len) -> List[float]:
        """
        This method turn list of windows back to ts suitable for visualization

        Args:
            windows (List[Window]): list of windows
            demanded_len (_type_): len of ts

        Returns:
            List[float]: time series
        """
        output_ts: List[float] = []
        for window in windows:
            for _ in range(0, self.step):
                output_ts.append(window.vector_result)
        for _ in range(len(output_ts), demanded_len):
            output_ts.append(0)
        return output_ts

    def _run_detector(self) -> None:
        self.windows = []
        self.windows = self._split_data_to_windows(
            self.data_object.time_series_data,
            self.data_object.get_len_of_dataset(),
            list(self.data_object.time_series_data.keys())
        )

        for j in range(len(self.windows)):
            self.windows[j].get_vector_predict()
        self.data_object.test_vector_ts = \
            self._turn_windows_to_ts(
                self.windows, self.data_object.get_len_of_dataset()
            )
        self.data_object.test_vector_ts = NormalizeData(
            self.data_object.test_vector_ts
        )

    def _print_logs(self, log_message: str) -> None:
        if self.print_logs:
            print(f"---[{time_now()}] {self.name}: {log_message}")



from typing import List, Mapping


class AnomalyZone:
    """
        + This class contains all data of one anomaly +
        Be careful, this class uses both for dataset objects of anomalies and for 
        anomalies that we are looking at!
    """
    def __init__(self, start: int = None, end: int = None) -> None:
        # Common fields
        # start and end counts of anomaly on data
        self.start: int = start
        self.end: int = end
        # parts of data that contains in this anomaly in form of 
        # an Mapping[str, list]
        self.data: Mapping[str, list] = {}
        # Additional time series that was generated for anomaly
        # Max and min time series that formed by choosing max and min of all 
        # data time seris en every point of anomaly
        self.min_data:  List[float] = None
        self.max_data:  List[float] = None
        # distance data - time seres that creates from distances from max and min
        # additional ts's in each point of anomaly
        self.distance_data: List[float] = None
        # ts like upper one, but cutted by threshold in Simple Detector element
        self.distance_data_trans: List[float] = None
        # additional ts that are max of average absolute deviation of all ts's in main data
        self.average_absolute_deviation: List[float] = None
        self.average_absolute_deviation_transformed: List[float] = None
        self.additional_mean_ts: List[float] = None
        # vector of features for searching anomalie's heaviness
        self.features_for_heaviness_clusterization: List[float] = []
        # vector of features for comparion anomalie to dataset of anomalies
        self.features_for_datasets: List[float] = []
        # vector of features for searching anomalie's type
        self.features_for_type_clusterization: List[float] = []

        # Type, description and heaviness of anomaly
        self.type: str = None
        self.comment: str = "None"
        self.heaviness: int = 0

        # For adataset usage - type of anomaly
        self.anomaly_class: int = None
        # For adataset usage - names of features
        self.features: List[str] = []
        # ???
        self.data_for_dataset: dict  = {}
        # +++ Threshold for <self.distance_data_trans> field +++
        self.threshold = 0
        # Fields for visualisation 
        # still in progress...
        self.x: float = 0
        self.y: float = 0
        # coordinates for visualisation of clusters 
        # during search of heaviness of animaly
        self.x_heaviness: float = 0
        self.y_heaviness: float = 0
        # coordinates for visualisation of clusters 
        # during search of type of animaly
        self.x_type: float = 0
        self.y_type: float = 0

        # predicted information
        # type and heaviness of anomaly that was get by comparing this anomaly
        # to dataset of anomalies 
        self.dataset_type: int = 0
        self.dataset_heaviness: int = 0
        self.dataset_comment: str = ""
        # type and heaviness of anomaly that was get by clusterizations
        self.cluster_type: int = 0
        self.cluster_heaviness: int = 0
        self.cluster_comment: str = ""
        # resulted type and heaviness of anomaly 
        self.predicted_type: int = 0
        self.predicted_heaviness: int = 0
        self.predicted_comment: str = ""


    def get_start(self) -> int:
        return self.start
        
    def get_end(self) -> int:
        return self.end
    
    



import logging
import matplotlib.pyplot as plt


class BenchmarkTester:
    """A class that performs model testing on selected datasets."""

    def __init__(self):
        self.logger = logging.getLogger('BenchmarkTester')
        self.results = []

    def test(self, model, dataset):
        """Perform model testing on a given dataset."""
        self.logger.info('Testing model on dataset...')
        result = model.evaluate(dataset)
        self.results.append(result)
        self.logger.info('Test complete.')
        return result


class ResultLogger:
    """A class that logs and saves test results."""

    def __init__(self, log_file):
        self.log_file = log_file
        self.logger = logging.getLogger('ResultLogger')

    def log(self, model_tester):
        """Log the test results from a ModelTester instance."""
        self.logger.info('Logging test results...')
        for result in model_tester.results:
            self.logger.info('Test result: %s', result)
        self.logger.info('Results logged.')


class ResultVisualizer:
    """A class that visualizes the progress of testing and its results."""

    def __init__(self):
        self.logger = logging.getLogger('ResultVisualizer')

    def visualize(self, model_tester):
        """Visualize the test results from a ModelTester instance."""
        self.logger.info('Visualizing test results...')
        plt.plot(model_tester.results)
        plt.title('Model Test Results')
        plt.xlabel('Test #')
        plt.ylabel('Result')
        plt.show()
        self.logger.info('Test results visualized.')


import logging
import os
from typing import Union

import numpy as np
import yaml

from fedot_ind.api.utils.checkers_collections import ParameterCheck
from fedot_ind.api.utils.hp_generator_collection import GeneratorParams
from fedot_ind.core.architecture.settings.pipeline_factory import FeatureGenerator
from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
from fedot_ind.core.architecture.utils.utils import PROJECT_PATH
from fedot_ind.core.models.BaseExtractor import BaseExtractor


class YamlReader:
    """
    Class for reading the config file for the experiment.

    Attributes:
        config_dict: dictionary with the parameters of the experiment.
        experiment_check: class for checking the correctness of the parameters.
    """

    def __init__(self):

        self.config_dict = None
        self.logger = logging.getLogger(self.__class__.__name__)
        self.experiment_check = ParameterCheck()
        self.experiment_dict = None

    def read_yaml_config(self,
                         # config_path: str,
                         # direct_path: bool = False,
                         # return_dict: bool = False
                         ) -> Union[None, dict]:
        """Read yaml config from directory as dictionary file.

        Args:
            config_path: path to the config file.
            direct_path: flag that indicates whether to use direct path to config file or read it from framework
            directory. config_path: name of the config file.
            return_dict: flag that indicates whether to return the dictionary with the parameters of the experiment.

        Returns:
            None

        """
        # if direct_path:
        #     path = config_path
        # else:
        #     path = os.path.join(PROJECT_PATH, config_path)
        #
        # with open(path, "r") as input_stream:
        #     config_dict = yaml.safe_load(input_stream)
        #     if 'path_to_config' in list(config_dict.keys()):
        #         config_path = config_dict['path_to_config']
        #         path = os.path.join(PROJECT_PATH, config_path)
        #         with open(path, "r") as input_stream:
        #             config_dict_template = yaml.safe_load(input_stream)
        #         config_dict = {**config_dict_template, **config_dict}
        #         del config_dict['path_to_config']
        #
        #     if 'baseline' not in config_dict.keys():
        #         config_dict['baseline'] = None
        #
        # if return_dict:
        #     return config_dict

        return dict(task='ts_classification',
                    dataset=None,
                    strategy='fedot_preset',
                    model_params={'problem': 'classification',
                                  'seed': 42,
                                  'timeout': 15,
                                  'max_depth': 10,
                                  'max_arity': 4,
                                  'cv_folds': 2,
                                  'logging_level': 20,
                                  'n_jobs': -1}
                    )

    def init_experiment_setup(self, **kwargs) -> dict:
        """Initializes the experiment setup with provided config file or dictionary.

        Args:
            kwargs: parameters of the experiment.

        Returns:
            Dictionary with the parameters of the experiment.

        """

        # base_config_path = os.path.join(PROJECT_PATH, 'api', 'config.yaml')
        # self.experiment_dict = self.read_yaml_config(config_path=base_config_path,
        #                                              return_dict=True)
        self.experiment_dict = self.read_yaml_config()

        fedot_config = {}
        industrial_config = {k: v if k not in ['timeout', 'n_jobs', 'logging_level'] else fedot_config.update({k: v})
                             for k, v in kwargs.items()}

        self.experiment_dict.update(**industrial_config)
        self.experiment_dict['model_params'].update(**fedot_config)

        self.experiment_dict['generator_class'] = self._get_generator_class()

        self.__report_experiment_setup(self.experiment_dict)

        return self.experiment_dict

    def _get_generator_class(self) -> Union[BaseExtractor, None]:
        """Support method that combines the name of the generator with the parameters from the config file.

        Returns:
            Class of the feature generator.

        """
        generator = self.experiment_dict['strategy']
        if generator is None:
            return None
        elif generator == 'fedot_preset':
            return None
        else:
            if generator.startswith('ensemble'):
                dict_of_generators = {}
                generators_to_ensemble = generator.split(': ')[1].split(' ')
                for gen in generators_to_ensemble:
                    single_gen_class = self._extract_generator_class(gen)
                    dict_of_generators[gen] = single_gen_class
                ensemble_gen_class = FeatureGenerator['ensemble'].value(list_of_generators=dict_of_generators)
                self.feature_generator = 'ensemble'
                return ensemble_gen_class

            feature_gen_class = self._extract_generator_class(generator)
            return feature_gen_class

    def _extract_generator_class(self, generator):
        feature_gen_model = FeatureGenerator[generator].value
        feature_gen_params = GeneratorParams[generator].value

        for param in feature_gen_params:
            feature_gen_params[param] = self.experiment_dict.get(param, feature_gen_params[param])

        feature_gen_class = feature_gen_model(feature_gen_params)
        return feature_gen_class

    def __report_experiment_setup(self, experiment_dict):

        self.logger.info(f'''Experiment setup:
        dataset - {experiment_dict['dataset']},
        strategy - {experiment_dict['strategy']},
        use_cache - {experiment_dict['use_cache']},
        n_jobs - {experiment_dict['model_params']['n_jobs']},
        timeout - {experiment_dict['model_params']['timeout']}''')


class DataReader:
    """
    Class for reading train and test data from the dataset. Exploits ``DataLoader`` class to read or download data
    from the UCR time series archive.

    """

    def __init__(self):

        self.logger = logging.getLogger(self.__class__.__name__)

    def read(self, dataset_name: str):

        # load data
        train_data, test_data = DataLoader(dataset_name).load_data()
        if train_data is None:
            self.logger.error(f'Some problem with {dataset_name} data. Skip it')
            return None, None, None
        else:
            n_classes = len(np.unique(train_data[1]))
            self.logger.info(f'Loaded data from {dataset_name} local data folder')
            self.logger.info(f'{n_classes} classes detected')
            return train_data, test_data, n_classes



from multiprocessing import Pool
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from fedot.core.operations.operation_parameters import OperationParameters
from tqdm import tqdm

from fedot_ind.api.utils.checkers_collections import DataCheck
from fedot_ind.core.metrics.metrics_implementation import ParetoMetrics
from fedot_ind.core.models.BaseExtractor import BaseExtractor
from fedot_ind.core.operation.decomposition.SpectrumDecomposition import SpectrumDecomposer
from fedot_ind.core.operation.transformation.data.eigen import combine_eigenvectors
from fedot_ind.core.operation.transformation.extraction.statistical import StatFeaturesExtractor
from fedot_ind.core.operation.transformation.regularization.spectrum import sv_to_explained_variance_ratio


class SSAExtractor(BaseExtractor):
    """Class responsible for spectral feature generator experiment.

    Args:
        window_sizes: dict of window sizes for SSA algorithm.
        window_mode: flag for window mode. If True, SSA algorithm will be applied to each window of time series.
        use_cache: flag for cache usage. If True, SSA algorithm will be applied to each window of time series.

    Attributes:
        aggregator (StatFeaturesExtractor): class for statistical features extraction.
        spectrum_extractor (SpectrumDecomposer): class for SSA algorithm.
        pareto_front (ParetoMetrics): class for pareto front calculation.
        vis_flag (bool): flag for visualization.
        rank_hyper (int): ...
        train_feats (pd.DataFrame): extracted features for train data.
        test_feats (pd.DataFrame): extracted features for test data.

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)

        self.combine_eigenvectors = params.get('combine_eigenvectors')
        self.correlation_level = params.get('correlation_level')
        self.use_cache = params.get('use_cache')
        self.window_sizes = params.get('window_sizes')
        self.window_mode = params.get('window_mode')
        self.spectral_hyperparams = params.get('spectral_hyperparams')
        self.__init_spectral_hyperparams()

        self.aggregator = StatFeaturesExtractor()
        self.spectrum_extractor = SpectrumDecomposer
        self.pareto_front = ParetoMetrics()
        self.datacheck = DataCheck()

        self.vis_flag = False
        self.rank_hyper = None
        self.train_feats = None
        self.test_feats = None
        self.n_components = None

    def __init_spectral_hyperparams(self):
        if self.spectral_hyperparams is not None:
            for k, v in self.spectral_hyperparams.items():
                setattr(self, k, v)

    @staticmethod
    def visualise(eigenvectors_list):

        n_rows = max((1, round(eigenvectors_list[0].shape[1] / 5)))

        if n_rows < 4:
            plot_area = 'small'
        elif 5 < n_rows < 9:
            plot_area = 'mid'
        else:
            plot_area = 'big'

        plot_dict = {'small': (15, 5),
                     'mid': (30, 10),
                     'big': (40, 20)}

        fig_size = plot_dict[plot_area]

        for idx, df in enumerate(eigenvectors_list):
            df.plot(subplots=True,
                    figsize=fig_size,
                    legend=None,
                    layout=(n_rows, df.shape[1]))

            plt.tight_layout()
            plt.show()

    def _ts_chunk_function(self, ts_data: pd.DataFrame) -> list:

        ts = self.check_for_nan(ts_data)
        specter = self.spectrum_extractor(time_series=ts, window_length=self.current_window)

        TS_comps, Sigma, rank, X_elem, V = specter.decompose(
            rank_hyper=self.rank_hyper)
        explained_variance, n_components = sv_to_explained_variance_ratio(Sigma, rank)

        if self.combine_eigenvectors:
            components_df = combine_eigenvectors(TS_comps, rank, self.correlation_level)
        else:
            components_df = specter.components_to_df(TS_comps, rank)

        return [components_df, n_components, explained_variance]

    def generate_vector_from_ts(self, ts_frame):
        ts_samples_count = ts_frame.shape[0]
        n_processes = self.n_processes
        self.logger.info(f'Number of processes: {n_processes}')

        with Pool(n_processes) as p:
            components_and_vectors = list(tqdm(p.imap(self._ts_chunk_function,
                                                      ts_frame.values),
                                               total=ts_samples_count,
                                               desc='Feature Generation. TS processed',
                                               unit=' ts',
                                               colour='black'
                                               )
                                          )
        self.logger.info(f'Number of time series processed: {ts_samples_count}')
        return components_and_vectors

    # @time_it
    def get_features(self, ts_data: pd.DataFrame, dataset_name: str = None, target: np.ndarray = None) -> pd.DataFrame:
        self.logger.info('Spectral features extraction started')
        if self.current_window is None:
            self._choose_best_window_size(ts_data, dataset_name=dataset_name)
            aggregation_df = self.train_feats
        else:
            eigenvectors_list_test = ts_data
            self.eigenvectors_list_test = self.check_rank_consistency(eigenvectors_list_test)
            aggregation_df = self.generate_features_from_ts(eigenvectors_list_test, window_mode=self.window_mode)
            aggregation_df = aggregation_df[self.train_feats.columns]

            for col in aggregation_df.columns:
                aggregation_df[col].fillna(value=aggregation_df[col].mean(), inplace=True)

        self.logger.info('Spectral features extraction finished')
        return aggregation_df

    def generate_features_from_ts(self, eigenvectors_list: list, window_mode: bool = False) -> pd.DataFrame:
        eigenvectors_list = list(map(lambda x: self.datacheck.check_data(x, return_df=True), eigenvectors_list))
        if self.window_mode:
            gen = self.aggregator.create_baseline_features
            function_for_stat_features = lambda x: self.apply_window_for_stat_feature(ts_data=x.T,
                                                                                      feature_generator=gen,
                                                                                      window_size=self.current_window)
            concat_function = lambda x: pd.concat(x, axis=1)

            list_with_stat_features_on_interval = list(map(function_for_stat_features, eigenvectors_list))
            aggregation_df = list(map(concat_function, list_with_stat_features_on_interval))
        else:
            aggregation_df = list(map(lambda x: self.aggregator.create_baseline_features(x.T), eigenvectors_list))

        components_names = aggregation_df[0].index.values
        columns_names = aggregation_df[0].columns.values

        aggregation_df = pd.concat([pd.DataFrame(x.values.ravel()) for x in aggregation_df], axis=1)
        aggregation_df.columns = range(len(aggregation_df.columns))
        aggregation_df = aggregation_df.T

        new_column_names = []
        for number_of_component in components_names:
            new_column_names.extend([f'{x}_for_component: {number_of_component}' for x in columns_names])

        aggregation_df.columns = new_column_names

        return aggregation_df

    def _choose_best_window_size(self, x_train: pd.DataFrame, dataset_name: str) -> pd.DataFrame:
        """Chooses the best window for feature extraction.

        Args:
            x_train: train data.
            dataset_name: name of dataset.

        Returns:
            Extracted features for train data.

        """
        self.logger.info('Choose best window size for feature extraction')
        metric_list = []
        n_comp_list = []
        eigen_list = []

        if self.window_sizes == 'auto':
            ts_length = x_train.shape[1]
            window_list = list(map(lambda x: round(ts_length / x), [10, 5, 3]))
        else:
            window_list = self.window_sizes

        for window_length in window_list:
            self.logger.info(f'Generate features for window length - {window_length}')
            self.current_window = window_length

            eigenvectors_and_rank = self.generate_vector_from_ts(x_train.sample(frac=0.5))

            rank_list = [x[1] for x in eigenvectors_and_rank]

            explained_dispersion = [x[2] for x in eigenvectors_and_rank]
            median_dispersion = np.median(explained_dispersion)
            self.explained_dispersion = round(float(median_dispersion))

            self.n_components = int(np.median(rank_list))

            eigenvectors_list = [x[0].iloc[:, :self.n_components] for x in eigenvectors_and_rank]

            self.logger.info(f'Every eigenvector with impact less then 1 % percent was eliminated. '
                             f'{self.explained_dispersion} % of explained dispersion '
                             f'obtained by first - {self.n_components} components.')
            signal_compression = 100 * (1 - (self.n_components / window_length))
            explained_dispersion = self.explained_dispersion
            metrics = [signal_compression, explained_dispersion]
            metric_list.append(metrics)

            eigen_list.append(eigenvectors_list)
            n_comp_list.append(self.n_components)
            if self.n_components > 15:
                self.logger.info(f'SSA method find {self.n_components} PCT.'
                                 f'This is mean that SSA method does not find effective low rank structure for '
                                 f'this signal.')
                break

        index_of_window = self.pareto_front.pareto_metric_list(np.array(metric_list))
        index_of_window = np.where(index_of_window == True)[0][0]
        self.current_window = window_list[index_of_window]
        self.logger.info(f'Best window length - {self.current_window} selected.')

        eigenvectors_list = eigen_list[index_of_window]
        self.rank_hyper = int(np.round(np.median([x.shape[1] for x in eigenvectors_list])))
        eigenvectors_and_rank = self.generate_vector_from_ts(x_train)
        eigenvectors_list_train = [x[0].iloc[:, :self.rank_hyper] for x in eigenvectors_and_rank]
        self.eigenvectors_list_train = self.check_rank_consistency(eigenvectors_list_train)
        self.train_feats = self.generate_features_from_ts(self.eigenvectors_list_train, window_mode=self.window_mode)
        for col in self.train_feats.columns:
            self.train_feats[col].fillna(value=self.train_feats[col].mean(), inplace=True)
        self.train_feats = self.delete_col_by_var(self.train_feats)
        self.n_components = n_comp_list[index_of_window]
        self.logger.info(f'Window length = {self.current_window} was chosen')

        return self.train_feats

    def __create_mask(self, eigenvectors_list_train):
        mask = [x.shape[1] < self.rank_hyper for x in eigenvectors_list_train]
        invalid_idx = [i for i, x in enumerate(mask) if x]
        return invalid_idx

    def check_rank_consistency(self, eigenvectors_list, rank=None):
        if rank is None:
            rank = self.rank_hyper

        invalid_idx = self.__create_mask(eigenvectors_list)
        while len(invalid_idx) != 0:
            eigenvectors_list = self.__fill_empty_eigenvectors(invalid_idx, eigenvectors_list, rank)
            invalid_idx = self.__create_mask(eigenvectors_list)
            if len(invalid_idx) == 0:
                break
        return eigenvectors_list

    def __fill_empty_eigenvectors(self, invalid_idx, eigenvectors_list, rank):
        for idx in invalid_idx:
            invalid_sample = eigenvectors_list[idx]
            missed_col = rank - invalid_sample.shape[1]
            last_number_of_component = invalid_sample.columns.values[-1]
            if missed_col == 1:
                eigenvectors_list[idx][last_number_of_component + missed_col] = 0
            else:
                for number_of_missed_component in range(1, missed_col):
                    eigenvectors_list[idx][last_number_of_component + number_of_missed_component] = 0

        return eigenvectors_list



import copy
import os
import time
from typing import Dict, List

import torch
from torch.utils.tensorboard import SummaryWriter

from fedot_ind.core.metrics.loss.svd_loss import HoyerLoss, OrthogonalLoss
from fedot_ind.core.models.cnn.decomposed_conv import DecomposedConv2d
from fedot_ind.core.models.cnn.sfp_models import SFP_MODELS
from fedot_ind.core.operation.optimization.sfp_tools import prune_resnet_state_dict, zerolize_filters
from fedot_ind.core.operation.optimization.svd_tools import decompose_module, energy_threshold_pruning


class GeneralizedStructureOptimization:
    """Generalized class for model structure optimization.

    Args:
        experimenter: An instance of the experimenter class, e.g.
            ``ClassificationExperimenter``.
    """

    def __init__(
        self,
        experimenter,
    ):
        self.exp = experimenter
        self.losses = {}

    def optimize_during_training(self) -> None:
        """Have to implement optimization method applied after train loop every epoch."""
        pass

    def final_optimize(self) -> None:
        """Have to implement optimization method applied once after training."""
        pass


class SVDOptimization(GeneralizedStructureOptimization):
    """Singular value decomposition for model structure optimization.

    Args:
        experimenter: An instance of the experimenter class, e.g.
            ``ClassificationExperimenter``.
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
        hoer_loss_factor: The hyperparameter by which the Hoyer loss function is
            multiplied.
        orthogonal_loss_factor: The hyperparameter by which the orthogonal loss
            function is multiplied.
        energy_thresholds: List of pruning hyperparameters.
        finetuning_epochs: Number of fine-tuning epochs.
    """

    def __init__(
        self,
        experimenter,
        decomposing_mode: str,
        hoer_loss_factor: float,
        orthogonal_loss_factor: float,
        energy_thresholds: List[float],
        finetuning_epochs: int,
    ) -> None:
        super().__init__(
            experimenter=experimenter,
        )
        decompose_module(experimenter.get_optimizable_module(), decomposing_mode)
        print(f"SVD decomposed size: {self.exp.size_of_model():.2f} MB")
        self.losses['hoer_loss'] = HoyerLoss(hoer_loss_factor)
        self.losses['orthogonal_loss'] = OrthogonalLoss(
            self.exp.device, orthogonal_loss_factor
        )
        self.energy_thresholds = energy_thresholds
        self.finetuning_epochs = finetuning_epochs
        self.exp.name += (
            f"_SVD_{decomposing_mode}_O-{orthogonal_loss_factor:.1f}"
            f"_H-{hoer_loss_factor:.6f}"
        )

    def final_optimize(self) -> None:
        """Apply optimization after training."""
        p_writer = SummaryWriter(
            os.path.join(self.exp.summary_path, self.exp.name, 'pruned')
        )
        ft_writer = SummaryWriter(
            os.path.join(self.exp.summary_path, self.exp.name, 'fine-tuned')
        )
        self.losses['hoer_loss'] = HoyerLoss(factor=0)
        self.exp.summary_per_class = False
        self.exp.load_model_state_dict()
        default_model = copy.deepcopy(self.exp.model)

        for e in self.energy_thresholds:
            int_e = int(e * 100000)
            self.exp.model = copy.deepcopy(default_model)
            start = time.time()
            self.prune_model(e)
            pruning_time = time.time() - start
            p_writer.add_scalar('pruning/pruning_time', pruning_time, int_e)
            self.optimization_summary(e=int_e, writer=p_writer)
            self.exp.finetune(
                num_epochs=self.finetuning_epochs,
                name=f"fine-tuning_e_{e}"
            )
            self.optimization_summary(e=int_e, writer=ft_writer)
            self.exp.save_model(name=f"fine-tuned_e_{e}")

    def prune_model(self, energy_threshold) -> None:
        """Prune the model weights to the energy_threshold.

        Args:
            energy_threshold: pruning hyperparameter, the lower the threshold, the more
        singular values will be pruned.
        """
        for module in self.exp.get_optimizable_module().modules():
            if isinstance(module, DecomposedConv2d):
                energy_threshold_pruning(conv=module, energy_threshold=energy_threshold)

    def optimization_summary(self, e: int, writer: SummaryWriter) -> None:
        """Validate model and write scores.

        Args:
            e: ``energy_threshold`` as integer for writing scores.
            writer: SummaryWriter object for writing scores.
        """
        val_scores = self.exp.val_loop()
        val_scores['size'] = self.exp.size_of_model()
        val_scores['n_params'] = self.exp.number_of_params()
        for key, score in val_scores.items():
            key = key.split('/')[-1]
            writer.add_scalar(f'pruning/{key}', score, e)


class SFPOptimization(GeneralizedStructureOptimization):
    """Soft filter pruning for model structure optimization.

    Args:
        experimenter: An instance of the experimenter class, e.g.
            ``ClassificationExperimenter``.
        pruning_ratio: Pruning hyperparameter, percentage of pruned filters.
        finetuning_epochs: Number of fine-tuning epochs.
    """

    def __init__(
            self,
            experimenter,
            pruning_ratio: float,
            finetuning_epochs: int,
    ) -> None:
        super().__init__(
            experimenter=experimenter,
        )
        self.pruning_ratio = pruning_ratio
        self.finetuning_epochs = finetuning_epochs
        self.exp.name += f"_SFP_P-{pruning_ratio:.2f}"

    def optimize_during_training(self) -> None:
        """ Apply optimization after train loop every epoch."""
        for module in self.exp.get_optimizable_module().modules():
            if isinstance(module, torch.nn.Conv2d):
                zerolize_filters(conv=module, pruning_ratio=self.pruning_ratio)

    def final_optimize(self) -> None:
        """Apply optimization after training."""
        self.exp.load_model_state_dict()
        self.exp.summary_per_class = False
        results = {'default_scores': self.optimization_summary()}

        pruned_sd, input_size, output_size = prune_resnet_state_dict(
            self.exp.model.state_dict()
        )
        self.exp.model = SFP_MODELS[self.exp.optimizable_module_name](
            num_classes=self.exp.num_classes,
            input_size=input_size,
            output_size=output_size,
            pruning_ratio=self.pruning_ratio
        )
        self.exp.model.load_state_dict(pruned_sd)
        self.exp.model.to(self.exp.device)
        results['pruned_scores'] = self.optimization_summary()
        self.exp.save_model(name='pruned')

        self.exp.finetune(num_epochs=self.finetuning_epochs)
        results['finetuned_scores'] = self.optimization_summary()
        self.exp.save_model(name='fine-tuned')

        for k in results['default_scores']:
            print(
                f"{k}: {results['default_scores'][k]:.3f}, "
                f"{results['pruned_scores'][k]:.3f}, "
                f"{results['finetuned_scores'][k]:.3f}"
            )
        torch.save(results, os.path.join(
            self.exp.summary_path, self.exp.name, 'results.pt'
        ))

    def optimization_summary(self) -> Dict:
        """Validate model and return scores."""
        scores = self.exp.val_loop()
        scores['size'] = self.exp.size_of_model()
        scores['n_params'] = self.exp.number_of_params()
        return scores


OPTIMIZATIONS = {
    'none': GeneralizedStructureOptimization,
    'SVD': SVDOptimization,
    'SFP': SFPOptimization,
}



from itertools import repeat
from multiprocessing import Pool
from typing import Optional

from fedot.core.operations.operation_parameters import OperationParameters
from tqdm import tqdm

from fedot_ind.core.metrics.metrics_implementation import *
from fedot_ind.core.models.signal.WindowedFeaturesExtractor import WindowedFeatureExtractor
from fedot_ind.core.operation.transformation.extraction.statistical import StatFeaturesExtractor


class SignalExtractor(WindowedFeatureExtractor):
    """Class responsible for wavelet feature generator experiment.
    Args:
        wavelet_types: list of wavelet types to be used in experiment. Defined in Config_Classification.yaml.
        use_cache: flag to use cache or not. Defined in Config_Classification.yaml
    Attributes:
        ts_samples_count (int): number of samples in time series
        aggregator (StatFeaturesExtractor): class to aggregate features
        wavelet_extractor (WaveletExtractor): class to extract wavelet features
        wavelet (str): current wavelet type
        vis_flag (bool): flag to visualize or not
        train_feats (pd.DataFrame): train features
        test_feats (pd.DataFrame): test features
        dict_of_methods (dict): dictionary of methods to extract features
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.ts_samples_count = None
        self.aggregator = StatFeaturesExtractor()
        self.wavelet_extractor = WaveletExtractor

        self.wavelet = params.get('wavelet')
        self.vis_flag = False
        self.train_feats = None
        self.test_feats = None
        self.dict_of_methods = {'Peaks': self._method_of_peaks,
                                'AC': self._method_of_AC}

    def _method_of_peaks(self, specter):
        threshold_range = [1, 3, 5, 7, 9]
        high_freq, low_freq = specter.decompose_signal()

        hf_lambda_peaks = lambda x: len(specter.detect_peaks(high_freq, mph=x + 1))
        hf_lambda_names = lambda x: 'HF_peaks_higher_than_{}'.format(x + 1)
        hf_lambda_knn = lambda x: len(specter.detect_peaks(high_freq, mpd=x))
        hf_lambda_knn_names = lambda x: 'HF_nearest_peaks_at_distance_{}'.format(x)

        lf_lambda_peaks = lambda x: len(specter.detect_peaks(high_freq, mph=x + 1, valley=True))
        lf_lambda_names = lambda x: 'LF_peaks_higher_than_{}'.format(x + 1)
        lf_lambda_knn = lambda x: len(specter.detect_peaks(high_freq, mpd=x))
        lf_lambda_knn_names = lambda x: 'LF_nearest_peaks_at_distance_{}'.format(x)

        lambda_list = [
            hf_lambda_knn,
            lf_lambda_peaks,
            lf_lambda_knn]

        lambda_list_names = [
            hf_lambda_knn_names,
            lf_lambda_names,
            lf_lambda_knn_names]

        features = list(map(hf_lambda_peaks, threshold_range))
        features_names = list(map(hf_lambda_names, threshold_range))
        for lambda_method, lambda_name in zip(lambda_list, lambda_list_names):
            features.extend(list(map(lambda_method, threshold_range)))
            features_names.extend(list(map(lambda_name, threshold_range)))

        feature_df = pd.DataFrame(data=features)
        feature_df = feature_df.T
        feature_df.columns = features_names
        return feature_df

    def _method_of_AC(self, specter, level: int = 3):
        high_freq, low_freq = specter.decompose_signal()
        hf_AC_features = specter.generate_features_from_AC(HF=high_freq,
                                                           LF=low_freq,
                                                           level=level)

        feature_df = pd.concat(hf_AC_features, axis=1)
        return feature_df

    def _ts_chunk_function(self, ts,
                           method_name: str = 'AC'):

        ts = self.check_for_nan(ts)
        specter = self.wavelet_extractor(time_series=ts, wavelet_name=self.wavelet)
        feature_df = self.dict_of_methods[method_name](specter)

        return feature_df

    def generate_vector_from_ts(self, ts_frame: pd.DataFrame, method_name: str = 'AC') -> list:
        """Generate vector from time series.
        Args:
            ts_frame (pd.DataFrame): time series to be transformed.
            method_name (str): method to be used for transformation.
        Returns:
            list: list of components and vectors.
        """
        ts_samples_count = ts_frame.shape[0]
        n_processes = self.n_processes
        with Pool(n_processes) as p:
            components_and_vectors = list(tqdm(p.starmap(self._ts_chunk_function,
                                                         zip(ts_frame, repeat(method_name))),
                                               total=ts_samples_count,
                                               desc='Feature Generation. TS processed',
                                               unit=' ts',
                                               colour='black'
                                               )
                                          )

        return components_and_vectors

    def generate_features_from_ts(self, ts_data: pd.DataFrame, dataset_name: str = None) -> pd.DataFrame:
        if not self.wavelet:
            train_feats = self._choose_best_wavelet(ts_data)
            self.train_feats = train_feats
            return self.train_feats
        else:
            test_feats = self.generate_vector_from_ts(ts_data)
            test_feats = pd.concat(test_feats)
            test_feats.index = list(range(len(test_feats)))
            self.test_feats = test_feats
        return self.test_feats



import sys
from copy import deepcopy
from functools import partial
from math import log

import numpy as np
from numpy import dot, eye, zeros
from sklearn.preprocessing import MinMaxScaler

from fedot_ind.core.architecture.experiment.TimeSeriesClassifier import TimeSeriesClassifier
from fedot_ind.core.models.detection.probalistic.sigma import MerweScaledSigmaPoints
from fedot_ind.core.operation.transformation.data.hankel import get_x_y_pairs


def reshape_z(z, dim_z, ndim):
    """ensure z is a (dim_z, 1) shaped vector"""

    z = np.atleast_2d(z)
    if z.shape[1] == dim_z:
        z = z.T

    if z.shape != (dim_z, 1):
        raise ValueError(
            "z (shape {}) must be convertible to shape ({}, 1)".format(z.shape, dim_z)
        )

    if ndim == 1:
        z = z[:, 0]

    if ndim == 0:
        z = z[0, 0]

    return z


class AbstractKalmanFilter:
    def __init__(self, model_hyperparams: dict):
        self.model_hyperparams = model_hyperparams
        # Only computed only if requested via property
        self._log_likelihood = log(sys.float_info.min)
        self._likelihood = sys.float_info.min
        self._mahalanobis = None
        # self.state = zeros((self.input_dim, 1))  # state
        self.state = None  # state
        self.control_transition_matrix = None  # B
        self.state_transition_matrix = None  # F
        self.inv = np.linalg.inv

    def _init_kalman_params(self, input_data):
        self.input_dim = input_data.shape[0]
        self.output_dim = self.input_dim

        self._I = np.eye(self.input_dim)  # identity matrix
        self.uncertainty_covariance = 100. * eye(self.input_dim)  # P
        self.process_uncertainty = 3. * eye(self.input_dim)  # R

        self.measurement_function = eye(self.input_dim)  # H
        self.measurement_uncertainty = 5. * eye(self.input_dim)  # Q
        self.fading_memory_control = 1.  # alpha
        self.process_measurement_cross_correlation = np.zeros((self.input_dim, self.output_dim))  # M
        self.z = np.array([[None] * self.output_dim]).T

        # gain and residual are computed during the innovation step. We
        # save them so that in case you want to inspect them for various
        # purposes
        self.kalman_gain = np.zeros((self.input_dim, self.output_dim))  # K
        self.residual = zeros((self.output_dim, 1))
        self.system_uncertainty = np.zeros((self.output_dim, self.output_dim))  # S
        self.inversed_system_uncertainty = np.zeros((self.output_dim, self.output_dim))  # SI

    def _fit_model_dynamic(self, train_features: np.ndarray):
        self._init_kalman_params(train_features)
        self.state = train_features
        self.state_mean = np.mean(train_features, axis=1)
        self.uncertainty_covariance = np.cov(train_features)
        self.train_features, self.target = get_x_y_pairs(train=train_features, train_periods=1,
                                                         prediction_periods=1)
        self.measurement_function = MinMaxScaler(feature_range=(0, 1))
        self.measurement_function.fit(self.train_features)
        self.train_features = self.measurement_function.transform(self.train_features)
        self.target = self.measurement_function.transform(self.target)
        self.state_transition_matrix = TimeSeriesClassifier(model_hyperparams=self.model_hyperparams)
        self.state_transition_matrix.fit(train_features=self.train_features, train_target=self.target)

    def fit(self, train_features, fit_settings: dict = {'control vector': None}):
        self._fit_model_dynamic(train_features=train_features)
        # x = Fx + Bu
        if self.state is None:
            self.state = train_features
        if self.control_transition_matrix is not None and fit_settings['control vector'] is not None:
            self.state = dot(self.state_transition_matrix, self.state) + dot(self.control_transition_matrix,
                                                                             fit_settings['control vector'])
        else:
            self.state = dot(self.state_transition_matrix, self.state)

        # P = FPF' + Q
        self.uncertainty_covariance = self.fading_memory_control * dot(
            dot(self.state_transition_matrix, self.uncertainty_covariance),
            self.state_transition_matrix.T) + self.process_uncertainty

        # save prior
        self.state_prior = self.state.copy()
        self.uncertainty_covariance_prior = self.uncertainty_covariance.copy()
        return self.state, self.uncertainty_covariance

    def predict(self, test_features, R=None, H=None):
        """
        Add a new measurement (z) to the Kalman filter.
        If z is None, nothing is computed. However, x_post and P_post are
        updated with the prior (x_prior, P_prior), and self.z is set to None.
        Parameters
        ----------
        test_features : (self.output_dim, 1): array_like
            measurement for this update. z can be a scalar if self.output_dim is 1,
            otherwise it must be convertible to a column vector.
            If you pass in a value of H, z must be a column vector the
            of the correct size.
        R : np.array, scalar, or None
            Optionally provide R to override the measurement noise for this
            one call, otherwise  self.R will be used.
        H : np.array, or None
            Optionally provide H to override the measurement function for this
            one call, otherwise self.H will be used.
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if test_features is None:
            self.z = np.array([[None] * self.output_dim]).T
            self.state_post = self.state.copy()
            self.uncertainty_covariance_post = self.uncertainty_covariance.copy()
            self.y = zeros((self.output_dim, 1))
            return

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.residual = test_features - self.state_transition_matrix.predict(self.state)

        # common subexpression for speed
        PHT = dot(self.uncertainty_covariance, H.T)

        # S = HPH' + R
        # project system uncertainty into measurement space
        self.system_uncertainty = dot(self.state_transition_matrix, PHT) + self.measurement_uncertainty
        self.inversed_system_uncertainty = self.inv(self.system_uncertainty)
        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.kalman_gain = dot(PHT, self.inversed_system_uncertainty)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.state = self.state + dot(self.kalman_gain, self.residual)

        # P = (I-KH)P(I-KH)' + KRK'
        # This is more numerically stable
        # and works for non-optimal K vs the equation
        # P = (I-KH)P usually seen in the literature.

        I_KH = self._I - dot(self.kalman_gain, self.measurement_function)
        self.uncertainty_covariance = dot(dot(I_KH, self.uncertainty_covariance), I_KH.T) + dot(dot(
            self.kalman_gain, self.measurement_uncertainty), self.kalman_gain.T)

        # save measurement and posterior state
        self.measurement = deepcopy(test_features)
        self.state_post = self.state.copy()
        self.uncertainty_covariance_post = self.uncertainty_covariance.copy()

        return self.state, self.uncertainty_covariance

    def residual_of(self, z):
        """
        Returns the residual for the given measurement (z). Does not alter
        the state of the filter.
        """
        z = reshape_z(z, self.self.output_dim, self.state.ndim)
        return z - dot(self.H, self.state_prior)

    def measurement_of_state(self, x):
        """
        Helper function that converts a state into a measurement.
        Parameters
        ----------
        x : np.array
            kalman state vector
        Returns
        -------
        z : (self.output_dim, 1): array_like
            measurement for this update. z can be a scalar if self.output_dim is 1,
            otherwise it must be convertible to a column vector.
        """

        return dot(self.measurement_function, x)


class UnscentedKalmanFilter(AbstractKalmanFilter):
    def __init__(self, model_hyperparams: dict):
        super().__init__(model_hyperparams)
        self.sigma_points = partial(MerweScaledSigmaPoints, alpha=.1, beta=2., kappa=-1)

    def fit(self, train_features, fit_settings: dict = {'control vector': None}):
        self._fit_model_dynamic(train_features=train_features)
        # calculate sigma points for given mean and covariance
        self.sigma_distribution = self.sigma_points(self.input_dim)
        self.sigmas_f = zeros((self.sigma_distribution.num_sigmas(), self.input_dim))
        self.sigmas_h = zeros((self.sigma_distribution.num_sigmas(), self.output_dim))

    def update(self):
        r"""
        Performs the predict step of the UKF. On return, self.state and
        self.P contain the predicted state (x) and covariance (P). '
        Important: this MUST be called before update() is called for the first
        time.
        """

        sigmas = self.sigma_distribution.sigma_points(self.state_mean, self.uncertainty_covariance)
        sigmas = self.measurement_function.transform(sigmas)
        for i, s in enumerate(sigmas):
            if len(s.shape) < 2:
                s = s.reshape(1, -1)
            self.sigmas_f[i] = self.state_transition_matrix.predict(s)
        self.sigmas_f = self.measurement_function.inverse_transform(self.sigmas_f)

        # pass sigmas through the unscented transform to compute prior
        self.state_mean, self.uncertainty_covariance = self.unscented_transform(sigmas=self.sigmas_f,
                                                                                Wm=self.sigma_distribution.Wm,
                                                                                Wc=self.sigma_distribution.Wc,
                                                                                noise_cov=self.process_uncertainty,
                                                                                mean_fn=None,
                                                                                residual_fn=np.subtract)

        # update sigma points to reflect the new variance of the points
        self.sigmas_f = self.sigma_distribution.sigma_points(self.state_mean, self.uncertainty_covariance)

        # save prior
        self.state_prior = np.copy(self.state_mean)
        self.uncertainty_covariance_prior = np.copy(self.uncertainty_covariance)

    def _predict(self, test_features,
                 measurement_uncertainty=None,
                 measurement_function=None):
        """
        Add a new measurement (z) to the Kalman filter.
        If z is None, nothing is computed. However, x_post and P_post are
        updated with the prior (x_prior, P_prior), and self.z is set to None.
        Parameters
        ----------
        test_features : measurement for this predict.
        measurement_uncertainty : np.array, scalar, or callable
            Optionally provide R to override the measurement noise for this
            one call, otherwise  self.R will be used.
        measurement function : np.array, or callable
            Optionally provide H to override the measurement function for this
            one call, otherwise self.H will be used.
        """
        # update prior_mean and prior_covariance
        self.update()
        # pass prior sigmas through h(x) to get measurement sigmas
        # the shape of sigmas_h will vary if the shape of z varies, so
        # recreate each time
        if measurement_function is None:
            self.sigmas_h = self.sigmas_f
        else:
            self.sigmas_h = dot(self.measurement_function, self.sigmas_f)

        self.sigmas_h = np.atleast_2d(self.sigmas_h)

        # mean and covariance of prediction passed through unscented transform
        measurement_mean, self.system_uncertainty = self.unscented_transform(sigmas=self.sigmas_h,
                                                                             Wm=self.sigma_distribution.Wm,
                                                                             Wc=self.sigma_distribution.Wc,
                                                                             noise_cov=self.measurement_uncertainty,
                                                                             mean_fn=None,
                                                                             residual_fn=np.subtract)

        self.inversed_system_uncertainty = self.inv(self.system_uncertainty)

        # compute cross variance of the state and the measurements
        Pxz = self.cross_variance(self.state_mean, measurement_mean, self.sigmas_f, self.sigmas_h)

        self.kalman_gain = dot(Pxz, self.inversed_system_uncertainty)  # Kalman gain
        self.residual = np.subtract(test_features, measurement_mean.reshape(-1, 1))
        weighted_residual = dot(self.kalman_gain, self.residual)  # residual
        predicted_state = test_features + weighted_residual
        weighted_mean = np.mean(weighted_residual, axis=1)
        # update Gaussian state estimate (x, P)
        self.state_mean = np.add(self.state_mean, weighted_mean)
        self.uncertainty_covariance = self.uncertainty_covariance - dot(self.kalman_gain,
                                                                        dot(self.system_uncertainty,
                                                                            self.kalman_gain.T))

        # save measurement and posterior state
        self.measurement = deepcopy(test_features)
        self.state_post = self.state.copy()
        self.uncertainty_covariance_post = self.uncertainty_covariance.copy()

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None
        return self.residual, predicted_state

    def predict(self, test_features,
                measurement_uncertainty=None,
                measurement_function=None):

        if type(test_features) == list:
            list_of_residuals, list_of_states = [], []
            for window_slice in test_features:
                try:
                    current_residual, current_state = self._predict(window_slice,
                                                                    measurement_uncertainty,
                                                                    measurement_function)
                    list_of_residuals.append(current_residual)
                    list_of_states.append(current_state)
                except Exception:
                    _ = 1
            return list_of_residuals, list_of_states
        else:
            return self._predict(test_features, measurement_uncertainty, measurement_function)

    @staticmethod
    def unscented_transform(sigmas,
                            Wm,
                            Wc,
                            noise_cov=None,
                            mean_fn=None,
                            residual_fn=None):
        r"""
        Computes unscented transform of a set of sigma points and weights.
        returns the mean and covariance in a tuple.
        This works in conjunction with the UnscentedKalmanFilter class.
        Parameters
        ----------
        sigmas: ndarray, of size (n, 2n+1)
            2D array of sigma points.
        Wm : ndarray [# sigmas per dimension]
            Weights for the mean.
        Wc : ndarray [# sigmas per dimension]
            Weights for the covariance.
        noise_cov : ndarray, optional
            noise matrix added to the final computed covariance matrix.
        mean_fn : callable (sigma_points, weights), optional
            Function that computes the mean of the provided sigma points
            and weights. Use this if your state variable contains nonlinear
            values such as angles which cannot be summed.
            .. code-block:: Python
                def state_mean(sigmas, Wm):
                    x = np.zeros(3)
                    sum_sin, sum_cos = 0., 0.
                    for i in range(len(sigmas)):
                        s = sigmas[i]
                        x[0] += s[0] * Wm[i]
                        x[1] += s[1] * Wm[i]
                        sum_sin += sin(s[2])*Wm[i]
                        sum_cos += cos(s[2])*Wm[i]
                    x[2] = atan2(sum_sin, sum_cos)
                    return x
        residual_fn : callable (x, y), optional
            Function that computes the residual (difference) between x and y.
            You will have to supply this if your state variable cannot support
            subtraction, such as angles (359-1 degreees is 2, not 358). x and y
            are state vectors, not scalars.
            .. code-block:: Python
                def residual(a, b):
                    y = a[0] - b[0]
                    y = y % (2 * np.pi)
                    if y > np.pi:
                        y -= 2*np.pi
                    return y
        Returns
        -------
        x : ndarray [dimension]
            Mean of the sigma points after passing through the transform.
        P : ndarray
            covariance of the sigma points after passing through the transform.
        """

        kmax, n = sigmas.shape

        try:
            if mean_fn is None:
                # new mean is just the sum of the sigmas * weight
                x = np.dot(Wm, sigmas)  # dot = \Sigma^n_1 (W[k]*Xi[k])
            else:
                x = mean_fn(sigmas, Wm)
        except:
            print(sigmas)
            raise

        # new covariance is the sum of the outer product of the residuals
        # times the weights

        # this is the fast way to do this - see 'else' for the slow way
        if residual_fn is np.subtract or residual_fn is None:
            y = sigmas - x[np.newaxis, :]
            P = np.dot(y.T, np.dot(np.diag(Wc), y))
        else:
            P = np.zeros((n, n))
            for k in range(kmax):
                y = residual_fn(sigmas[k], x)
                P += Wc[k] * np.outer(y, y)

        if noise_cov is not None:
            P += noise_cov

        return x, P

    def cross_variance(self, x, z, sigmas_f, sigmas_h):
        """
        Compute cross variance of the state `x` and measurement `z`.
        """

        Pxz = zeros((sigmas_f.shape[1], sigmas_h.shape[1]))
        N = sigmas_f.shape[0]
        for i in range(N):
            dx = np.subtract(sigmas_f[i], x)
            dz = np.subtract(sigmas_h[i], z)
            Pxz += self.sigma_distribution.Wc[i] * np.outer(dx, dz)
        return Pxz



from typing import Optional

import torch
from torch.nn.modules import Module
from torch.nn.modules.conv import Conv2d

from fedot_ind.core.models.cnn.decomposed_conv import DecomposedConv2d


def energy_threshold_pruning(conv: DecomposedConv2d, energy_threshold: float) -> None:
    """Prune the weight matrices to the energy_threshold (in-place).

    Args:
        conv: The optimizable layer.
        energy_threshold: pruning hyperparameter must be in the range (0, 1].
        the lower the threshold, the more singular values will be pruned.

    Raises:
        Assertion Error: If ``conv.decomposing`` is False.
    """
    assert conv.decomposing, "for pruning, the model must be decomposed"
    assert 0 < energy_threshold <= 1, "energy_threshold must be in the range (0, 1]"
    S, indices = conv.S.sort()
    U = conv.U[:, indices]
    Vh = conv.Vh[indices, :]
    sum = (S ** 2).sum()
    threshold = energy_threshold * sum
    for i, s in enumerate(S):
        sum -= s ** 2
        if sum < threshold:
            conv.set_U_S_Vh(U[:, i:].clone(), S[i:].clone(), Vh[i:, :].clone())
            break


def decompose_module(model: Module, decomposing_mode: Optional[str] = None) -> None:
    """Replace Conv2d layers with DecomposedConv2d layers in module (in-place).

    Args:
        model: Decomposable module.
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
            If ``None`` replace layers without decomposition. Default: ``None``
    """
    for name, module in model.named_children():
        if len(list(module.children())) > 0:
            decompose_module(module, decomposing_mode=decomposing_mode)

        if isinstance(module, Conv2d):
            new_module = DecomposedConv2d(
                in_channels=module.in_channels,
                out_channels=module.out_channels,
                kernel_size=module.kernel_size,
                stride=module.stride,
                padding=module.padding,
                dilation=module.dilation,
                groups=module.groups,
                bias=(module.bias is not None),
                padding_mode=module.padding_mode,
                decomposing=False,
            )
            new_module.load_state_dict(module.state_dict())
            if decomposing_mode is not None:
                new_module.decompose(decomposing_mode=decomposing_mode)
            setattr(model, name, new_module)


def _load_svd_params(model, state_dict, prefix='') -> None:
    """Loads state_dict to DecomposedConv2d layers in model."""
    for name, module in model.named_children():
        if len(list(module.children())) > 0:
            _load_svd_params(module, state_dict, prefix=f'{prefix}{name}.')

        if isinstance(module, DecomposedConv2d):
            module.set_U_S_Vh(
                u=state_dict[f'{prefix}{name}.U'],
                s=state_dict[f'{prefix}{name}.S'],
                vh=state_dict[f'{prefix}{name}.Vh']
            )


def load_svd_state_dict(
        model: Module,
        decomposing_mode: str,
        state_dict_path: str
) -> None:
    """Loads SVD state_dict to model.

    Args:
        model: An instance of the base model.
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
        state_dict_path: Path to state_dict file.
    """
    state_dict = torch.load(state_dict_path, map_location='cpu')
    decompose_module(model=model, decomposing_mode=decomposing_mode)
    _load_svd_params(model, state_dict)
    model.load_state_dict(state_dict)



import logging
import os
from typing import Union

import pandas as pd

from fedot_ind.core.architecture.utils.utils import PROJECT_PATH

DS_INFO_PATH = os.path.join(PROJECT_PATH, 'core', 'architecture', 'postprocessing', 'ucr_datasets.json')


class ResultsPicker:
    """Class for parsing results of experiments. It parses all experiments in ``results_of_experiments``
    folder or user-specified folder and creates dataframe table that availible for further analysis.

    Args:
        path (str): path to folder with experiments. Default is ``None`` and it means that path
                    is ``results_of_experiments``.
        launch_type (str or int): number of launch to be extracted. Default is ``max`` and it means that the best launch
                                  will be extracted.

    Examples:
        >>> from fedot_ind.core.architecture.postprocessing.results_picker import ResultsPicker
        >>> collector = ResultsPicker(path='to_your_results_folder', launch_type='max')
        >>> metrics_df = parser.run(get_metrics_df=True)
        >>> metrics_df.to_csv('metrics.csv')

    """

    def __init__(self, path: str = None, launch_type: Union[str, int] = 'max'):
        self.exp_path = self.__get_results_path(path)
        self.launch_type = launch_type
        self.logger = logging.getLogger(self.__class__.__name__)


    def __get_results_path(self, path):
        if path:
            return path
        else:
            return os.path.join(PROJECT_PATH, 'results_of_experiments')

    def run(self, get_metrics_df: bool = False, add_info: bool = False):
        """
        Base method for parsing results of experiments.

        Returns:
            Table with results of experiments.

        """

        proba_dict, metric_dict = self.get_metrics_and_proba()

        if get_metrics_df:
            if add_info:
                metrics_df = self._create_metrics_df(metric_dict)
                datasets_info = self.get_datasets_info()
                return pd.merge(metrics_df, datasets_info, how='left', on='dataset')
            return self._create_metrics_df(metric_dict)

        return proba_dict, metric_dict

    def _create_metrics_df(self, metric_dict):
        columns = ['dataset', 'experiment']
        metrics_df = pd.DataFrame()
        for ds in metric_dict.keys():
            for exp in metric_dict[ds].keys():
                metrics = metric_dict[ds][exp].to_dict(orient='records')[0]
                metrics_df = metrics_df.append({'dataset': ds, 'experiment': exp,
                                                'f1': metrics.get('f1'), 'roc_auc': metrics.get('roc_auc'),
                                                'accuracy': metrics.get('accuracy'), 'precision': metrics.get('precision'),
                                                'logloss': metrics.get('logloss')}, ignore_index=True)

        metrics_df = pd.concat([metrics_df[['dataset', 'experiment']],
                                metrics_df[[col for col in metrics_df.columns if col not in columns]]], axis=1)
        return metrics_df

    def get_metrics_and_proba(self):
        experiments = self.list_dirs(self.exp_path)
        proba_dict = {}
        metric_dict = {}
        for exp in experiments:
            exp_path = os.path.join(self.exp_path, exp)
            ds_list, metrics_list, proba_list = self.read_exp_folder(exp_path)

            if ds_list is None:
                continue

            for metric, proba, dataset in zip(metrics_list, proba_list, ds_list):
                if dataset not in proba_dict.keys() and proba is not None:
                    proba_dict[dataset] = {}
                if dataset not in metric_dict.keys() and proba is not None:
                    metric_dict[dataset] = {}
                proba_dict[dataset].update({exp: proba})
                metric_dict[dataset].update({exp: metric})

        return proba_dict, metric_dict

    def read_exp_folder(self, folder):
        # datasets_path = os.path.join(self.exp_path, folder)
        datasets = self.list_dirs(folder)
        metrics_list = []
        proba_list = []
        for ds in datasets:
            ds_path = os.path.join(folder, ds)
            metrics, proba = self.read_ds_data(ds_path)
            metrics_list.append(metrics)
            proba_list.append(proba)

        return datasets, metrics_list, proba_list

    def read_ds_data(self, path):
        # if self.launch_type == 'max':
        #     best_launch = self.find_best_launch(path)
        # else:
        #     best_launch = self.launch_type
        # metrics_path = os.path.join(path, best_launch, 'test_results', 'metrics.csv')
        # proba_path = os.path.join(path, best_launch, ' 'test_results', 'metrics.csv')
        metrics_path = os.path.join(path, 'metrics.csv')
        proba_path = os.path.join(path, 'predicted_probs.csv')

        try:
            proba = pd.read_csv(proba_path, index_col=0)
            metrics = pd.read_csv(metrics_path, index_col=0)
        except FileNotFoundError:
            self.logger.error(f'File not found: {metrics_path}')
            return None, None
        if 'index' in metrics.columns:
            del metrics['index']
            metrics = metrics.T
            metrics = metrics.rename(columns=metrics.iloc[0])
            metrics = metrics[1:]

        return metrics, proba

    @staticmethod
    def list_dirs(path):
        """Function used instead of ``os.listdir()`` to get list of non-hidden directories.

        Args:
            path (str): Path to the directory.

        Returns:
            list: List of non-hidden directories.

        """
        path_list = []
        for f in os.listdir(path):
            # if not f.startswith('.'):
            if '.' not in f:
                path_list.append(f)
        return path_list

    @staticmethod
    def list_files(path):
        """Function used instead of ``os.listdir()`` to get list of non-hidden files.

        Args:
            path (str): Path to the directory.

        Returns:
            list: List of non-hidden files.

        """
        path_list = []
        for f in os.listdir(path):
            if os.path.isfile(path + '/' + f) and not f.startswith('.'):
                path_list.append(f)

        return path_list

    def find_best_launch(self, launch_folders):
        best_metric = 0
        launch = 1
        for _dir in self.list_dirs(launch_folders):
            if len(_dir) == 1:
                metric_path = os.path.join(launch_folders, str(_dir), 'test_results', 'metrics.csv')
                metrics = pd.read_csv(metric_path, index_col=0)
                if 'index' in metrics.columns:
                    del metrics['index']
                    metrics = metrics.T
                    metrics = metrics.rename(columns=metrics.iloc[0])
                    metrics = metrics[1:]
                metric_sum = metrics['roc_auc'].values[0] + metrics['f1'].values[0]
                if metric_sum > best_metric:
                    best_metric = metric_sum
                    launch = _dir
        return launch

    def get_datasets_info(self):

        table = pd.read_json(DS_INFO_PATH)

        table = table.drop([col for col in table.columns if len(col) == 1] + ['Dataset_id'], axis=1)
        table.columns = list(map(str.lower, table.columns))
        table.type = table.type.str.lower()

        return table


# Example of usage:
if __name__ == '__main__':
    parser = ResultsPicker()
    results_table = parser.run(get_metrics_df=True, add_info=True)
    results_table.to_csv('results_4.csv', index=False)



import pathlib
from typing import Optional, Tuple

from golem.core.optimisers.genetic.evaluation import MultiprocessingDispatcher, OptionalEvalResult
from golem.core.optimisers.graph import OptGraph

from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels


class IndustrialDispatcher(MultiprocessingDispatcher):
    def evaluate_single(self, graph: OptGraph, uid_of_individual: str, with_time_limit: bool = True,
                        cache_key: Optional[str] = None,
                        logs_initializer: Optional[Tuple[int, pathlib.Path]] = None) -> OptionalEvalResult:
        # we should do that due to this is copy of initial process, and it does not have industrial models in itself
        if self._n_jobs != 1:
            with IndustrialModels():
                eval_res = super().evaluate_single(graph, uid_of_individual, with_time_limit, cache_key, logs_initializer)
        else:
            eval_res = super().evaluate_single(graph, uid_of_individual, with_time_limit, cache_key, logs_initializer)
        return eval_res



from fedot_ind.api.main import FedotIndustrial

if __name__ == "__main__":
    import os.path
    from fedot_ind.core.architecture.postprocessing.results_picker import ResultsPicker

    datasets = ['ItalyPowerDemand',
                'UMD'
                ]

    for dataset_name in datasets:

        industrial = FedotIndustrial(task='ts_classification',
                                     dataset=dataset_name,
                                     strategy='fedot_preset',
                                     # strategy='statistical',
                                     # strategy='wavelet',
                                     use_cache=True,
                                     timeout=1,
                                     n_jobs=2,
                                     window_sizes='auto',
                                     logging_level=20,
                                     output_folder=None)

        train_data, test_data, _ = industrial.reader.read(dataset_name=dataset_name)
        model = industrial.fit(train_features=train_data[0], train_target=train_data[1])

        labels = industrial.predict(test_features=test_data[0],
                                    test_target=test_data[1])

        probs = industrial.predict_proba(test_features=test_data[0],
                                         test_target=test_data[1])
        metric = industrial.get_metrics(target=test_data[1],
                                        metric_names=['f1', 'roc_auc'])

        for pred, kind in zip([labels, probs], ['labels', 'probs']):
            industrial.save_predict(predicted_data=pred, kind=kind)

        industrial.save_metrics(metrics=metric)

    results_path = os.path.abspath('../../results_of_experiments')
    picker = ResultsPicker(path=results_path)
    proba_dict, metric_dict = picker.run()
    from fedot_ind.core.ensemble.static.RankEnsembler import RankEnsemble

    ensembler = RankEnsemble(dataset_name='ItalyPowerDemand',
                             proba_dict=proba_dict,
                             metric_dict=metric_dict)
    ensembler.ensemble()



import logging
from abc import abstractmethod

dict_of_dataset = dict
dict_of_win_list = dict


class BaseEnsemble:
    """
    Abstract class responsible for models ensemble
        :param models_dict: dict that consists of 'feature_generator_method': model_class pairs
    """

    def __init__(self, dataset_name: str = None,
                 feature_generator_dict: dict = None):
        self.feature_generator_dict = feature_generator_dict
        self.dataset_name = dataset_name
        self.logger = logging.getLogger(self.__class__.__name__)
        self.metric_dict = None
        self.prediction_proba_dict = None

    @abstractmethod
    def ensemble(self) -> dict:
        raise NotImplementedError

    def create_proba_and_metrics_dicts(self, modelling_results: dict) -> (dict, dict):
        """
        Method for creating dictionary with structure {'ModelName': [tensor with class probs]}
        and dictionary with structure {'ModelName':[metric values]}

        Args:
            modelling_results: dict of results from modelling

        Returns:
            tuple of dicts
        """
        prediction_proba_dict = {}
        metric_dict = {}
        for generator, generator_results in modelling_results.items():
            if not modelling_results[generator]:
                continue
            best_launch = self._select_best_launch(generator_results)
            prediction_proba = generator_results[best_launch]['class_probability']
            metrics = generator_results[best_launch]['metrics']

            prediction_proba_dict.update({generator: prediction_proba})
            metric_dict.update({generator: metrics})

        return prediction_proba_dict, metric_dict

    def _select_best_launch(self, generator_results: dict) -> dict:
        """Method for selecting best launch from modelling results

        """
        best_metric = 0
        launch = 1
        for launch_num in generator_results.keys():
            metrics = generator_results[launch_num]['metrics']
            metric_sum = metrics['roc_auc'] + metrics['f1']
            if metric_sum > best_metric:
                best_metric = metric_sum
                launch = launch_num

        return launch



import json
import os
from typing import Tuple, Callable, Dict

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset
from tqdm import tqdm


def collate_fn(batch):
    return tuple(zip(*batch))


class COCODataset(Dataset):
    """Class-loader for COCO json.

    Args:
        images_path: Image folder path.
        json_path: Json file path.
        transform: A function/transform that takes in an PIL image and returns a
            transformed version.
    """

    def __init__(
        self,
        images_path: str,
        json_path: str,
        transform: Callable,
    ) -> None:
        self.transform = transform
        self.classes = {}
        self.samples = []

        with open(json_path) as f:
            data = json.load(f)

        for category in data["categories"]:
            self.classes[category["id"]] = category["name"]

        samples = {}
        for image in data["images"]:
            samples[image["id"]] = {
                "image": os.path.join(images_path, image["file_name"]),
                "area": [],
                "iscrowd": [],
                "labels": [],
                "boxes": [],
            }

        for annotation in tqdm(data["annotations"]):
            if annotation["area"] > 0:
                bbox = np.array(annotation["bbox"])
                bbox[2:] += bbox[:2]  # x, y, w, h -> x1, y1, x2, y2
                samples[annotation["image_id"]]["labels"].append(
                    annotation["category_id"]
                )
                samples[annotation["image_id"]]["boxes"].append(bbox)
                samples[annotation["image_id"]]["area"].append(annotation["area"])
                samples[annotation["image_id"]]["iscrowd"].append(annotation["iscrowd"])

        for sample in samples.values():
            if len(sample["labels"]) > 0:
                self.samples.append(sample)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Returns a sample from a dataset.

        Args:
            idx: Index of sample.

        Returns:
            A tuple ``(image, targets)``, where image is image tensor,
                and targets is dict with keys: ``'boxes'``, ``'labels'``,
                ``'image_id'``, ``'area'``, ``'iscrowd'``.
        """
        sample = self.samples[idx]
        image = Image.open(sample["image"]).convert("RGB")
        image = self.transform(image)
        target = {
            "boxes": torch.tensor(np.stack(sample["boxes"]), dtype=torch.float32),
            "labels": torch.tensor(sample["labels"], dtype=torch.int64),
            "image_id": torch.tensor([idx]),
            "area": torch.tensor(sample["area"], dtype=torch.float32),
            "iscrowd": torch.tensor(sample["iscrowd"], dtype=torch.int64),
        }
        return image, target

    def __len__(self) -> int:
        """Return length of dataset"""
        return len(self.samples)



from matplotlib import pyplot as plt

plt.rcParams['figure.figsize'] = (10, 8)
plt.rcParams['font.size'] = 14
plt.rcParams['image.cmap'] = 'plasma'
plt.rcParams['axes.linewidth'] = 2

cols = plt.get_cmap('tab10').colors

def plot_2d(m, title=""):
    plt.imshow(m)
    plt.xticks([])
    plt.yticks([])
    plt.title(title)

def plot_wcorr(minimum=None, maximum=None):
    """Plots the w-correlation matrix for the decomposed time series.

    """
    if minimum is None:
        minimum = 0
    if maximum is None:
        maximum = d

    if Wcorr is None:
        calc_wcorr()

    ax = plt.imshow(Wcorr)
    plt.xlabel(r"$\tilde{F}_i$")
    plt.ylabel(r"$\tilde{F}_j$")
    plt.colorbar(ax.colorbar, fraction=0.045)
    ax.colorbar.set_label("$W_{i,j}$")
    plt.clim(0, 1)

    # For plotting purposes:
    if maximum == d:
        max_range = d - 1
    else:
        max_range = maximum

    plt.xlim(minimum - 0.5, max_range + 0.5)
    plt.ylim(max_range + 0.5, minimum - 0.5)


from torch import nn
from torchvision.models import resnet18, resnet34, resnet50, resnet101, resnet152, ResNet


def resnet18_one_channel(**kwargs) -> ResNet:
    """ResNet18 for one input channel"""
    model = resnet18(**kwargs)
    model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
    return model


def resnet34_one_channel(**kwargs) -> ResNet:
    """ResNet34 for one input channel"""
    model = resnet34(**kwargs)
    model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
    return model


def resnet50_one_channel(**kwargs) -> ResNet:
    """ResNet50 for one input channel"""
    model = resnet50(**kwargs)
    model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
    return model


def resnet101_one_channel(**kwargs) -> ResNet:
    """ResNet101 for one input channel"""
    model = resnet101(**kwargs)
    model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
    return model


def resnet152_one_channel(**kwargs) -> ResNet:
    """ResNet152 for one input channel"""
    model = resnet152(**kwargs)
    model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
    return model


CLF_MODELS = {
    'ResNet18': resnet18,
    'ResNet34': resnet34,
    'ResNet50': resnet50,
    'ResNet101': resnet101,
    'ResNet152': resnet152,
    'ResNet18one': resnet18_one_channel,
    'ResNet34one': resnet34_one_channel,
    'ResNet50one': resnet50_one_channel,
    'ResNet101one': resnet101_one_channel,
    'ResNet152one': resnet152_one_channel,
}

CLF_MODELS_ONE_CHANNEL = {
    'ResNet18one': resnet18_one_channel,
    'ResNet34one': resnet34_one_channel,
    'ResNet50one': resnet50_one_channel,
    'ResNet101one': resnet101_one_channel,
    'ResNet152one': resnet152_one_channel,
}



from typing import Any, List, Optional, Type, Union, Dict

import torch
import torch.nn as nn
from torch import Tensor


def conv3x3(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=1,
        bias=False,
    )


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        planes: int,
        input_size: int,
        output_size: int,
        pruning_ratio: float,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
    ) -> None:
        super().__init__()
        norm_layer = nn.BatchNorm2d
        planes = planes - int(pruning_ratio * planes)
        self.conv1 = conv3x3(input_size, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, output_size)
        self.bn2 = norm_layer(output_size)
        self.downsample = downsample
        self.stride = stride
        if downsample is not None:
            self.register_buffer('indexes', torch.zeros(planes, dtype=torch.int))
        else:
            self.register_buffer('indexes', torch.zeros(input_size, dtype=torch.int))

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out.index_add_(1, self.indexes, identity)
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):

    expansion: int = 4

    def __init__(
        self,
        planes: int,
        input_size: int,
        output_size: int,
        pruning_ratio: float,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
    ) -> None:
        super().__init__()
        if downsample is not None:
            s = planes * self.expansion - int(planes * self.expansion * pruning_ratio)
            self.register_buffer('indexes', torch.zeros(s, dtype=torch.int))
        else:
            self.register_buffer('indexes', torch.zeros(input_size, dtype=torch.int))
        norm_layer = nn.BatchNorm2d
        planes = planes - int(pruning_ratio * planes)
        self.conv1 = conv1x1(input_size, planes)
        self.bn1 = norm_layer(planes)
        self.conv2 = conv3x3(planes, planes, stride)
        self.bn2 = norm_layer(planes)
        self.conv3 = conv1x1(planes, output_size)
        self.bn3 = norm_layer(output_size)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out.index_add_(1, self.indexes, identity)
        out = self.relu(out)

        return out


class ResNetSFP(nn.Module):
    """Pruned ResNet for soft filter pruning optimization.

    Args:
        block: ``'BasicBlock'`` or ``'Bottleneck'``.
        layers: Number of blocks on each layer.
        input_size: Input size of each block.
        output_size: Output size of each block.
        pruning_ratio: Pruning hyperparameter, percentage of pruned filters.
        num_classes: Number of classes.
    """
    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        input_size: Dict[str, List[int]],
        output_size: Dict[str, List[int]],
        pruning_ratio: float,
        num_classes: int,
    ) -> None:
        super().__init__()
        self.pr = pruning_ratio
        self.inplanes = 64
        in_size = self.inplanes - int(self.inplanes * self.pr)
        self.conv1 = nn.Conv2d(3, in_size, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(in_size)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(
            block=block,
            planes=64,
            blocks=layers[0],
            input_size=input_size['layer1'],
            output_size=output_size['layer1'],
        )
        self.layer2 = self._make_layer(
            block=block,
            planes=128,
            blocks=layers[1],
            input_size=input_size['layer2'],
            output_size=output_size['layer2'],
            stride=2)
        self.layer3 = self._make_layer(
            block=block,
            planes=256,
            blocks=layers[2],
            input_size=input_size['layer3'],
            output_size=output_size['layer3'],
            stride=2)
        self.layer4 = self._make_layer(
            block=block,
            planes=512,
            blocks=layers[3],
            input_size=input_size['layer4'],
            output_size=output_size['layer4'],
            stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(output_size['layer4'][-1], num_classes)

    def _make_layer(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        input_size: List[int],
        output_size: List[int],
        planes: int,
        blocks: int,
        stride: int = 1,
    ) -> nn.Sequential:
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            out_size = planes * block.expansion
            out_size = out_size - int(out_size * self.pr)
            downsample = nn.Sequential(
                conv1x1(input_size[0], out_size, stride),
                nn.BatchNorm2d(out_size),
            )
        layers = []
        layers.append(
            block(
                planes=planes,
                input_size=input_size[0],
                output_size=output_size[0],
                pruning_ratio=self.pr,
                stride=stride,
                downsample=downsample
            )
        )
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(
                planes=planes,
                input_size=input_size[i],
                output_size=output_size[i],
                pruning_ratio=self.pr
            ))
        return nn.Sequential(*layers)

    def _forward_impl(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)


def sfp_resnet18(**kwargs: Any) -> ResNetSFP:
    """Pruned ResNet-18."""
    return ResNetSFP(BasicBlock, [2, 2, 2, 2], **kwargs)


def sfp_resnet34(**kwargs: Any) -> ResNetSFP:
    """Pruned ResNet-34."""
    return ResNetSFP(BasicBlock, [3, 4, 6, 3], **kwargs)


def sfp_resnet50(**kwargs: Any) -> ResNetSFP:
    """Pruned ResNet-50."""
    return ResNetSFP(Bottleneck, [3, 4, 6, 3], **kwargs)


def sfp_resnet101(**kwargs: Any) -> ResNetSFP:
    """Pruned ResNet-101."""
    return ResNetSFP(Bottleneck, [3, 4, 23, 3], **kwargs)


def sfp_resnet152(**kwargs: Any) -> ResNetSFP:
    """Pruned ResNet-152."""
    return ResNetSFP(Bottleneck, [3, 8, 36, 3], **kwargs)


SFP_MODELS = {
    "ResNet18": sfp_resnet18,
    "ResNet34": sfp_resnet34,
    "ResNet50": sfp_resnet50,
    "ResNet101": sfp_resnet101,
    "ResNet152": sfp_resnet152,
}



import numpy as np
import pandas as pd
from MKLpy.algorithms import MEMO, CKA
from MKLpy.scheduler import ReduceOnWorsening
from MKLpy.callbacks import EarlyStopping
from fedot_ind.core.architecture.pipelines.classification import ClassificationPipelines
from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
from scipy.spatial.distance import pdist, squareform
from fedot_ind.core.architecture.settings.pipeline_factory import KernelFeatureGenerator


class KernelEnsembler(ClassificationPipelines):
    def __init__(self, train_data, test_data):
        super().__init__(train_data, test_data)
        self.n_classes = np.unique(self.train_target)
        if self.n_classes.shape[0] > 2:
            self.multiclass_strategy = 'ova'
            self.multiclass = True
        else:
            self.multiclass_strategy = 'ovr'
            self.multiclass = False
        self.kernel_params_list = {i.name: i.value for i in KernelFeatureGenerator}

    def __call__(self, pipeline_type: str = 'one_stage_kernel'):
        pipeline_dict = {'one_stage_kernel': self.__one_stage_kernel,
                         'two_stage_kernel': self.__two_stage_kernel
                         }
        return pipeline_dict[pipeline_type]

    def __convert_weights(self, kernel_model):
        kernels_weights_by_class = []
        if not self.multiclass:
            kernels_weights_by_class.append(kernel_model.solution.weights.cpu().detach().numpy())
        else:
            for n_class in self.n_classes:
                kernels_weights_by_class.append(kernel_model.solution[n_class].weights.cpu().detach().numpy())
        kernel_weight_matrix = pd.DataFrame(kernels_weights_by_class).apply(lambda x: abs(x))
        return kernel_weight_matrix

    def transform(self, kernel_params_dict: dict = None, feature_generator: str = None):
        feature_matrix = []
        if kernel_params_dict is None:
            kernel_params_dict = self.kernel_params_list
        if feature_generator is not None:
            kernel_params_dict = {feature_generator: kernel_params_dict[feature_generator]}

        for feature_generator, kernel_params in kernel_params_dict.items():
            for specified_params in kernel_params:
                feature_extractor, classificator, lambda_func_dict = self._init_pipeline_nodes(**specified_params)

                feature_matrix.append(self.get_feature_generator(input_data=self.train_features,
                                                                 steps=lambda_func_dict,
                                                                 pipeline_type='SpecifiedFeatureGeneratorTSC').value[0])
        return feature_matrix

    def __one_stage_kernel(self, kernel_params_dict: dict = None, feature_generator: str = None):
        feature_matrix = self.transform(kernel_params_dict, feature_generator)
        KLtr = [squareform(pdist(X=feature, metric='cosine')) for feature in feature_matrix]
        mkl = CKA(multiclass_strategy=self.multiclass_strategy).fit(KLtr, self.train_target)
        #mkl = FHeuristic(multiclass_strategy=self.multiclass_strategy).fit(KLtr, self.train_target)
        kernel_weight_matrix = self.__convert_weights(mkl)
        return kernel_weight_matrix

    def __two_stage_kernel(self, kernel_params_dict: dict = None, feature_generator: str = None):
        feature_matrix = self.transform(kernel_params_dict, feature_generator)
        KLtr = [squareform(pdist(X=feature, metric='cosine')) for feature in feature_matrix]
        earlystop = EarlyStopping(
            KLtr,
            self.train_target,  # validation data, KL is a validation kernels list
            patience=5,  # max number of acceptable negative steps
            cooldown=1,  # how ofter we run a measurement, 1 means every optimization step
            metric='roc_auc',  # the metric we monitor
        )

        mkl = MEMO(multiclass_strategy='ova',
                   max_iter=1000,
                   learning_rate=.1,
                   callbacks=[earlystop],
                   scheduler=ReduceOnWorsening()).fit(KLtr, self.train_target)
        kernel_weight_matrix = self.__convert_weights(mkl)
        return kernel_weight_matrix


if __name__ == '__main__':
    dataset_name = 'Earthquakes'
    train, test = DataLoader(dataset_name).load_data()
    kernels = KernelEnsembler(train_data=train, test_data=test)
    kernel_list = {'wavelet': [
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "mexh",
             'n_components': 2
         }},
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "morl",
             'n_components': 2
         }}]}
    set_of_fg = kernels('two_stage_kernel')(feature_generator='wavelet')

    # model_hyperparams = {
    #     'problem': 'classification',
    #     'seed': 42,
    #     'timeout': 4,
    #     'max_depth': 6,
    #     'max_arity': 3,
    #     'cv_folds': 3,
    #     'logging_level': 20,
    #     'n_jobs': 4
    # }
    # feature_hyperparams = {
    #     'window_mode': True,
    #     'window_size': 10
    # }
    # model_stats, prediction_stats = pipelines('SpecifiedFeatureGeneratorTSC')(feature_generator_type='statistical',
    #                                                                           model_hyperparams=model_hyperparams,
    #                                                                           feature_hyperparams=feature_hyperparams)
    # feature_hyperparams = {
    #     'wavelet': "morl"
    # }
    # model_wavelet, prediction_wavelet = pipelines('SpecifiedFeatureGeneratorTSC')(feature_generator_type='wavelet',
    #                                                                               model_hyperparams=model_hyperparams,
    #                                                                               feature_hyperparams=feature_hyperparams)

    # metrics = PerformanceAnalyzer().calculate_metrics(target=pipelines.test_target,
    #                                                   predicted_labels=prediction_stats['predicted_labels'],
    #                                                   predicted_probs=prediction_stats['predicted_probs'])



from typing import Union

import torch
from torch import Tensor
from torch.nn import Conv2d, Parameter
from torch.nn.common_types import _size_2_t

__all__ = ["DecomposedConv2d"]


class DecomposedConv2d(Conv2d):
    """Extends the Conv2d layer by implementing the singular value decomposition of
    the weight matrix.

    Args:
        in_channels: Number of channels in the input image
        out_channels: Number of channels produced by the convolution
        kernel_size: Size of the convolving kernel
        stride: Stride of the convolution. Default: ``1``
        padding: Padding added to all four sides of the input. Default: ``0``
        padding_mode: ``'zeros'``, ``'reflect'``, ``'replicate'`` or ``'circular'``.
            Default: ``'zeros'``
        decomposing: If ``True``, decomposes weights after initialization.
            Default: ``True``
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
            Default: ``'channel'``
        dilation: Spacing between kernel elements. Default: ``1``
        groups: Number of blocked connections from input channels to output channels.
            Default: ``1``
        bias: If ``True``, adds a learnable bias to the output. Default: ``True``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: Union[str, _size_2_t] = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = 'zeros',
        decomposing: bool = True,
        decomposing_mode: str = 'channel',
        device=None,
        dtype=None,
    ) -> None:

        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
            device,
            dtype,
        )

        n, c, w, h = self.weight.size()
        self.decomposing_modes_dict = {
            'channel': (n, c * w * h),
            'spatial': (n * w, c * h),
        }

        if decomposing:
            self.decompose(decomposing_mode)
        else:
            self.U = None
            self.S = None
            self.Vh = None
            self.decomposing = False

    def decompose(self, decomposing_mode: str) -> None:
        """Decomposes the weight matrix in singular value decomposition.
        Replaces the weights with U, S, Vh matrices such that weights = U * S * Vh.
        Args:
            decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
        Raises:
            ValueError: If ``decomposing_mode`` not in valid values.
        """
        if decomposing_mode not in self.decomposing_modes_dict.keys():
            raise ValueError(
                "decomposing_mode must be one of {}, but got decomposing_mode='{}'".format(
                    self.decomposing_modes_dict.keys(), decomposing_mode
                )
            )
        W = self.weight.view(self.decomposing_modes_dict[decomposing_mode])
        U, S, Vh = torch.linalg.svd(W, full_matrices=False)

        self.U = Parameter(U)
        self.S = Parameter(S)
        self.Vh = Parameter(Vh)
        self.register_parameter('weight', None)
        self.decomposing = True

    def compose(self) -> None:
        """Compose the weight matrix from singular value decomposition.
        Replaces U, S, Vh matrices with weights such that weights = U * S * Vh.
        """

        W = self.U @ torch.diag(self.S) @ self.Vh
        self.weight = Parameter(
            W.view(
                self.out_channels, self.in_channels // self.groups, *self.kernel_size
            )
        )

        self.register_parameter('U', None)
        self.register_parameter('S', None)
        self.register_parameter('Vh', None)
        self.decomposing = False

    def forward(self, input: Tensor) -> Tensor:

        if self.decomposing:
            W = self.U @ torch.diag(self.S) @ self.Vh
            return self._conv_forward(
                input,
                W.view(
                    self.out_channels,
                    self.in_channels // self.groups,
                    *self.kernel_size
                ),
                self.bias,
            )
        else:
            return self._conv_forward(input, self.weight, self.bias)

    def set_U_S_Vh(self, u: Tensor, s: Tensor, vh: Tensor) -> None:
        """Update U, S, Vh matrices.
        Raises:
            Assertion Error: If ``self.decomposing`` is False.
        """
        assert self.decomposing, "for setting U, S and Vh, the model must be decomposed"
        self.U = Parameter(u)
        self.S = Parameter(s)
        self.Vh = Parameter(vh)


import numpy as np

from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
from sklearn.metrics.pairwise import euclidean_distances


class SoftDTWLoss:
    def __init__(self, X, Y):
        self.X = X
        self.Y = Y

    def _soft_min_argmin(self, a, b, c, gamma):
        """Computes the soft min and argmin of (a, b, c).
        Args:
          a: scalar value.
          b: scalar value.
          c: scalar value.
        Returns:
          softmin, softargmin[0], softargmin[1], softargmin[2]
        """
        a /= -gamma
        b /= -gamma
        c /= -gamma

        max_val = max(max(a, b), c)

        sum_of_probs = 0
        exp_a = np.exp(a - max_val)
        exp_b = np.exp(b - max_val)
        exp_c = np.exp(c - max_val)
        sum_of_probs = exp_a + exp_b + exp_c
        softmin_value = -gamma * (np.log(sum_of_probs) + max_val)
        return softmin_value, exp_a, exp_b, exp_c
        # min_abc = min(a, min(b, c))
        # exp_a = np.exp(min_abc - a)
        # exp_b = np.exp(min_abc - b)
        # exp_c = np.exp(min_abc - c)
        # sum_of_probs = exp_a + exp_b + exp_c
        # exp_a /= sum_of_probs
        # exp_b /= sum_of_probs
        # exp_c /= sum_of_probs
        # val = min_abc - np.log(sum_of_probs)
        # return val, exp_a, exp_b, exp_c

    def _sdtw_C(self, cost_matrix, V, P, gamma):
        """SDTW dynamic programming recursion.
        Args:
          C: cost matrix (input).
          V: intermediate values (output).
          P: transition probability matrix (output).
        """
        size_X, size_Y = cost_matrix.shape

        for i in range(1, size_X + 1):
            for j in range(1, size_Y + 1):
                smin, P[i, j, 0], P[i, j, 1], P[i, j, 2] = \
                    self._soft_min_argmin(V[i - 1, j], V[i - 1, j - 1], V[i, j - 1], gamma=gamma)

                # The cost matrix C is indexed starting from 0.
                V[i, j] = cost_matrix[i - 1, j - 1] + smin
        return cost_matrix, V, P

    def sdtw_C(self, C, gamma=1.0, return_all=True):
        """Computes the soft-DTW value from a cost matrix C.
      Args:
        C: cost matrix, numpy array of shape (size_X, size_Y).
        gamma: regularization strength (scalar value).
        return_all: whether to return intermediate computations.
      Returns:
        sdtw_value if not return_all
        V (intermediate values), P (transition probability matrix) if return_all
      """
        size_X, size_Y = C.shape

        # # Handle regularization parameter 'gamma'.
        # if gamma != 1.0:
        #     C = C / gamma

        # Matrix containing the values of sdtw.
        V = np.zeros((size_X + 1, size_Y + 1))
        V[:, 0] = 1e10
        V[0, :] = 1e10
        V[0, 0] = 0

        # Tensor containing the probabilities of transition.
        P = np.zeros((size_X + 2, size_Y + 2, 3))

        C, V, P = self._sdtw_C(C, V, P, gamma)

        if return_all:
            return V, P
        else:
            return V[size_X, size_Y]

    def sdtw(self, gamma=1.0, return_all=False):
        """Computes the soft-DTW value from time series X and Y.
      The cost is assumed to be the squared Euclidean one.
      Args:
        X: time series, numpy array of shape (size_X, num_dim).
        Y: time series, numpy array of shape (size_Y, num_dim).
        gamma: regularization strength (scalar value).
        return_all: whether to return intermediate computations.
      Returns:
        sdtw_value if not return_all
        V (intermediate values), P (transition probability matrix) if return_all
      """
        cost_matrix = self.squared_euclidean_cost()
        return self.sdtw_C(cost_matrix, gamma=gamma, return_all=return_all)

    def squared_euclidean_cost(self):
        """Computes the squared Euclidean cost.
        """
        return euclidean_distances(self.X, self.Y, squared=True)


if __name__ == "__main__":
    # Two 3-dimensional time series of lengths 5 and 4, respectively.
    X = np.random.randn(5, 3)
    Y = np.random.randn(5, 3)
    train, test = DataLoader('ECG200').load_data()
    X = train[0].iloc[0:1, ].T.values
    Y = train[0].iloc[1:2, ].T.values
    metric = SoftDTWLoss(X=X, Y=Y)
    dtw_map, dtw_path = metric.sdtw(gamma=0.7)
    path = metric.compute_path(dtw_map)



import numpy as np
import scipy
from numpy.linalg import svd
from scipy.sparse.linalg import svds


def rq(A):
    n, m = A.shape()
    Q, R = np.linalg.qr(np.flipud(A).T, mode='complete')
    R = np.rot90(R.T, 2)
    Q = np.flipud(Q.T)
    if n > m:
        R = np.hstack((np.zeros((n, n - m)), R))
        Q = np.vstack((np.zeros((n - m, m)), Q))
    return R, Q


def tls(A, B):
    n = A.shape[1]
    if A.shape[0] != B.shape[0]:
        raise ValueError('Matrices are not conformant.')
    R1 = np.hstack((A, B))
    U, S, V = np.linalg.svd(R1)
    r = A.shape[1]
    R = rq(V[:, r:])
    Gamma = R[n:, n - r:]
    Z = R[:n, n - r:]
    Xhat = -np.dot(Z, np.linalg.inv(Gamma))
    return Xhat


def dmd_decompose():
    pass


def exact_dmd_decompose(X, Y, rank):
    Ux, Sx, Vx = svd(X)
    Ux = Ux[:, :rank]
    Sx = Sx[:rank]
    Sx = np.diag(Sx)
    Vx = Vx[:, :rank]
    # Project A onto the leading r principal components of X.
    Atilde = (Ux.T @ Y) @ Vx @ np.linalg.pinv(Sx)
    A = lambda v: np.dot(a=Ux, b=np.dot(a=Atilde, b=np.dot(a=Ux.T, b=v)))
    # Diagonalise linear operator
    eigen_vals, eigen_vectors = np.linalg.eig(Atilde)
    eigen_vals = np.diag(eigen_vals)
    # Approximation of operator eigenvectors
    eigen_vectors = Y @ Vx @ np.linalg.pinv(Sx) @ eigen_vectors / eigen_vals.T

    return A, eigen_vals, eigen_vectors


def orthogonal_dmd_decompose(X, Y, rank):
    Ux, _, _ = svd(X)
    Ux = Ux[:, :rank]
    # Project X (current state) and Y (future state) on leading components of X
    Yproj = Ux.T @ Y
    Xproj = Ux.T @ X
    # A_proj is constrained to be a unitary matrix and the minimization problem is argmin (A.T @ A = I) |Y-AX|_frob
    # The solution of A_proj is obtained by Schonemann A = Uyx,@ Vyx.T
    Uyx, _, Vyx = svd(Yproj @ Xproj.T)
    Aproj = Uyx @ Vyx.T
    A = lambda x: np.dot(a=Ux, b=np.dot(a=Aproj, b=np.dot(a=Ux.T, b=x)))
    # Diagonalise unitary operator
    eVals, eVecs = np.linalg.eig(Aproj)
    eigen_vals = np.diag(eVals)
    # Approximation of operator eigenvectors
    eigen_vectors = Ux @ eVecs
    return A, eigen_vals, eigen_vectors


def symmetric_decompose(X, Y, rank):
    Ux, S, V = np.linalg.svd(X)
    C = np.dot(Ux.T, np.dot(Y, V))
    C1 = C
    if rank is None:
        r = np.linalg.matrix_rank(X)
    Ux = Ux[:, :rank]
    Yf = np.zeros((rank, rank))
    for i in range(r):
        Yf[i, i] = np.real(C1[i, i]) / S[i, i]
        for j in range(i + 1, r):
            Yf[i, j] = (S[i, i] * np.conj(C1[j, i]) + S[j, j] * C1[i, j]) / (S[i, i] ** 2 + S[j, j] ** 2)
    Yf = Yf + Yf.T - np.diag(np.diag(np.real(Yf)))
    # elif method == 'skewsymmetric':
    #     for i in range(r):
    #         Yf[i, i] = 1j * np.imag(C1[i, i]) / S[i, i]
    #         for j in range(i + 1, r):
    #             Yf[i, j] = (-S[i, i] * np.conj(C1[j, i]) + S[j, j] * (C1[i, j])) / (S[i, i] ** 2 + S[j, j] ** 2)
    #     Yf = Yf - Yf.T - 1j * np.diag(np.diag(np.imag(Yf)))
    A = lambda v: np.dot(Ux, np.dot(Yf, np.dot(Ux.T, v)))
    return A


def hankel_decompose(X, Y, rank):
    nx, nt = X.shape()
    # J = np.eye(nx)
    J = np.fliplr(np.eye(nx))
    # Define the left matrix
    A = np.fft.fft(np.concatenate((np.eye(nx), np.zeros((nx, nx))), axis=1), axis=0) / np.sqrt(2 * nx)
    # Define the right matrix
    B = np.fft.fft(np.vstack((np.dot(J, X), np.zeros((nt, nx)))), axis=0) / np.sqrt(2 * nx)
    BtB = B.T @ B
    # Fast computation of A * A
    AAt = np.fft.ifft(np.fft.fft(np.concatenate((np.eye(nx), np.zeros((nx, nx))), axis=1)).T).T
    # Construct the RHS of the linear system
    y = np.diag(np.dot(np.dot(AAt.T, np.conj(Y)), B))
    # Construct the matrix for the linear system
    L = AAt @ BtB
    # Solve the linear system
    d = np.append(y[0:len(y) - 1] / L[0:len(L) - 1, 0:len(L) - 1], 0)
    # Convert the eigenvalues into the circulant matrix
    newA = np.fft.ifft(np.fft.fft(np.diag(d))).T
    # Extract the Toeplitz  matrix from the circulant matrix
    A = newA[0:nx, 0:nx] * J
    return A



from typing import Union

import numpy as np
import pandas as pd

from fedot_ind.core.architecture.settings.hyperparams import select_hyper_param

stat_methods = select_hyper_param('statistical_methods')
stat_methods_extra = select_hyper_param('statistical_methods_extra')
quantile_dict = {'q5_': 0.05,
                 'q25_': 0.25,
                 'q75_': 0.75,
                 'q95_': 0.95}


class StatFeaturesExtractor:
    """Class for generating statistical features for a given time series.

    """

    @staticmethod
    def create_baseline_features(feature_to_aggregation: Union[pd.DataFrame, np.ndarray]):
        stat_list = []
        column_name = []
        feature_to_aggregation = pd.DataFrame(feature_to_aggregation)
        if feature_to_aggregation.shape[0] != 1:
            feature_to_aggregation = feature_to_aggregation.T
        for method_name, method_func in stat_methods_extra.items():
            try:
                tmp = feature_to_aggregation.copy(deep=True)
            except Exception:
                tmp = feature_to_aggregation.copy()

            if method_name.startswith('q'):
                _ = []
                for idx, row in tmp.iterrows():
                    _.append(method_func(row, q=quantile_dict[method_name]))
                tmp = np.array(_)
                stat_list.append(tmp)
                column_name.append(method_name)
            elif method_name.startswith('l'):
                tmp = tmp.apply(method_func, axis=1)
                tmp = tmp.astype(int)
                stat_list.append(tmp.sum(axis=1).values)
                column_name.append(method_name)
            elif method_name.startswith('d'):
                tmp = tmp.apply(method_func, axis=1)
                stat_list.append(tmp.apply(np.mean).values)
                column_name.append(method_name + 'mean')
                stat_list.append(tmp.apply(np.min).values)
                column_name.append(method_name + 'min')
                stat_list.append(tmp.apply(np.max).values)
                column_name.append(method_name + 'max')
            else:
                stat_list.append(tmp.apply(method_func, axis=1).values)
                column_name.append(method_name)

            del tmp

        df_points_stat = pd.DataFrame(stat_list)
        df_points_stat = df_points_stat.T
        df_points_stat.columns = column_name

        return df_points_stat

    @staticmethod
    def create_features(feature_to_aggregation: Union[pd.DataFrame, np.ndarray]):
        stat_list = []
        column_name = []
        feature_to_aggregation = pd.DataFrame(feature_to_aggregation)
        for method_name, method_func in stat_methods.items():
            tmp = feature_to_aggregation.copy()

            if method_name.startswith('q'):
                for col in tmp.columns:
                    tmp[col] = method_func(tmp[col], q=quantile_dict[method_name])
                    tmp = tmp.drop_duplicates()
            else:
                tmp = pd.DataFrame(tmp.apply(method_func))
                tmp = tmp.T

            for feature in feature_to_aggregation.columns:
                column_name.append(method_name + str(feature))

            stat_list.append(tmp.values)

        df_points_stat = pd.DataFrame(np.concatenate(stat_list, axis=1))
        df_points_stat.columns = column_name
        return df_points_stat

    def _transform(self, X, intervals) -> np.ndarray:
        """
        Transform X for given intervals. Compute the mean, standard deviation and slope for given
        intervals of input data X.

        Args:
            X: input data
            intervals: list of intervals for which to compute the mean, standard deviation and slope

        Returns:
            Array of shape (len(X), 3 * len(intervals))
        """
        n_instances, _ = X.shape
        n_intervals, _ = intervals.shape
        transformed_x = np.empty(shape=(3 * n_intervals, n_instances), dtype=np.float32)
        for j in range(n_intervals):
            X_slice = X[:, intervals[j][0]: intervals[j][1]]
            means = np.mean(X_slice, axis=1)
            std_dev = np.std(X_slice, axis=1)
            slope = _slope(X_slice, axis=1)
            transformed_x[3 * j] = means
            transformed_x[3 * j + 1] = std_dev
            transformed_x[3 * j + 2] = slope

        return transformed_x.T

    @staticmethod
    def _get_intervals(n_intervals: int,
                       min_interval: int,
                       series_length: int,
                       rng) -> np.ndarray:
        """Generate random intervals for given parameters.

        Args:
            n_intervals: Number of intervals to generate
            min_interval: Minimum length of an interval
            series_length: Length of the time series
            rng: ...

        Returns:
            Array containing the intervals.

        """
        intervals = np.zeros((n_intervals, 2), dtype=int)
        for j in range(n_intervals):
            intervals[j][0] = rng.randint(series_length - min_interval)
            length = rng.randint(series_length - intervals[j][0] - 1)
            if length < min_interval:
                length = min_interval
            intervals[j][1] = intervals[j][0] + length
        return intervals




import numpy as np
from matplotlib import pyplot as plt


class DataProjection:
    """
    A class to project data into a different basis.

    Attributes:
        data (np.array): An array of data to be projected.
    """

    def __init__(self, data):
        """
        The constructor for the DataProjection class.

        Parameters:
            data (np.array): An array of data to be projected.
        """
        self.data = data

    def project_data(self, basis):
        """
        Projects the data into a different basis.

        Parameters:
            basis (np.array): The basis to project the data into.

        Returns:
            np.array: The projected data.
        """
        return np.dot(self.data, basis)

    def visualize_data(self):
        """
        Visualizes the data.
        """
        plt.scatter(self.data[:, 0], self.data[:, 1])
        plt.show()

    def visualize_projection(self, basis):
        """
        Visualizes the data projected into a different basis.

        Parameters:
            basis (np.array): The basis to project the data into.
        """
        projected_data = self.project_data(basis)
        plt.scatter(projected_data[:, 0], projected_data[:, 1])
        plt.show()

    def __add__(self, other):
        """
        Adds two DataProjection classes together.

        Parameters:
            other (DataProjection): The other DataProjection class to add.

        Returns:
            DataProjection: The combined DataProjection classes.
        """
        return DataProjection(self.data + other.data)

    def __sub__(self, other):
        """
        Subtracts two DataProjection classes.

        Parameters:
            other (DataProjection): The other DataProjection class to subtract.

        Returns:
            DataProjection: The subtracted DataProjection classes.
        """
        return DataProjection(self.data - other.data)

    def __mul__(self, other):
        """
        Multiplies two DataProjection classes.

        Parameters:
            other (DataProjection): The other DataProjection class to multiply.

        Returns:
            DataProjection: The multiplied DataProjection classes.
        """
        return DataProjection(self.data * other.data)



import logging

import numpy as np
import pandas as pd

from fedot_ind.core.architecture.postprocessing.Analyzer import PerformanceAnalyzer
from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
from fedot_ind.core.architecture.settings.hyperparams import select_hyper_param
from fedot_ind.core.ensemble.BaseEnsembler import BaseEnsemble


class RankEnsemble(BaseEnsemble):
    """Class responsible for the results of ensemble models
    by ranking them and recursively adding them to the final composite model.

    Args:
        dataset_name: name of dataset
        proba_dict: dictionary with prediction probabilities
        metric_dict: dictionary with metrics

    Attributes:

    """

    def __init__(self, dataset_name: str, proba_dict, metric_dict):
        super().__init__(dataset_name=dataset_name)
        self.proba_dict = proba_dict
        self.metric_dict = metric_dict

        self.performance_analyzer = PerformanceAnalyzer()
        self.experiment_results = {}
        self.logger = logging.getLogger(self.__class__.__name__)
        self.best_ensemble_metric = 0

        self.ensemble_strategy_dict = select_hyper_param('stat_methods_ensemble')
        self.ensemble_strategy = self.ensemble_strategy_dict.keys()

        self.strategy_exclude_list = ['WeightedEnsemble']

    def ensemble(self) -> dict:
        """Returns dictionary with ranking ensemble results

        The process of ensemble consists of 3 stages. At the first stage, a dictionary is created
        that contains the name of the model as a key and the best metric value for this dataset as a value.
        The second stage is the creation of a ranked list in the form of a dictionary (self.sorted_dict),
        also at this stage parameters such as the best model and the best value of the quality metric are determined,
        which are stored in the dictionary self.best_base_results. The third stage is iterative, in accordance
        with the assigned rank, adding models to a single composite model and ensemble their predictions.

        Returns:
            Fitted Fedot pipeline with baseline model

        """
        (_, _), (_, self.test_target) = DataLoader(self.dataset_name).load_data()

        prediction_proba_dict, metric_dict = self.proba_dict[self.dataset_name], self.metric_dict[self.dataset_name]

        model_rank_dict = self._create_models_rank_dict(prediction_proba_dict, metric_dict)
        self.best_base_results = self._sort_models(model_rank_dict)

        return self.__iterative_model_selection(prediction_proba_dict)

    def _deep_search_in_dict(self, obj, key):
        if key in obj:
            return obj[key]
        for k, v in obj.items():
            if isinstance(v, dict):
                item = self._deep_search_in_dict(v, key)
                if item is not None:
                    return item

    def _create_models_rank_dict(self, prediction_proba_dict, metric_dict):
        """
        Method that returns a dictionary with the best metric values of base models

        Returns:
            dictionary with structure {'ModelName': best metric values}

        """
        model_rank = {}
        # for model in self.metric_dict:
        for model in metric_dict:
            self.logger.info(f'BASE RESULT FOR MODEL - {model}'.center(50, '-'))
            # if len(self.prediction_proba_dict[model].columns) == 3:
            if prediction_proba_dict[model].shape[1] == 2:
                self.metric = 'roc_auc'
                _type = 'binary'
            else:
                self.metric = 'f1'
                _type = 'multiclass'
            self.logger.info(f'TYPE OF ML TASK - {_type}. Metric - {self.metric}'.center(50, '-'))
            current_metrics = metric_dict[model]
            if isinstance(current_metrics, pd.DataFrame):
                current_metrics = current_metrics.loc[0].to_dict()
            self.logger.info(current_metrics)
            model_rank.update({model: current_metrics[self.metric]})
        return model_rank

    def _sort_models(self, model_rank):
        """
        Method that returns sorted dictionary with models results ``

        Args:
            model_rank: dictionary with structure {'ModelName': 'best metric values'}

        Returns:
            sorted dictionary with structure {'Base_model': 'best metric values'}

        """

        self.sorted_dict = dict(sorted(model_rank.items(), key=lambda x: x[1], reverse=True))
        self.n_models = len(self.sorted_dict)

        best_base_model = list(self.sorted_dict)[0]
        best_metric = self.sorted_dict[best_base_model]

        self.logger.info(f'CURRENT BEST METRIC - {best_metric}. MODEL - {best_base_model}'.center(50, '-'))
        self.experiment_results.update({'Base_model': best_base_model, 'Base_metric': best_metric})
        return {'Base_model': best_base_model, 'Base_metric': best_metric}

    def __iterative_model_selection(self, prediction_proba_dict):
        """
        Method that iterative adding models to a single composite model and ensemble their predictions

        Returns:
            dictionary with structure {'Ensemble_models': 'best ensemble metric',
                                        'Base_model': 'best base model metric'}

        """
        for top_K_models in range(1, self.n_models):

            modelling_results_top = {k: v for k, v in prediction_proba_dict.items() if
                                     k in list(self.sorted_dict.keys())[:top_K_models + 1]}
            self.logger.info(f'Applying ensemble {self.ensemble_strategy} strategy for {top_K_models+1} models')

            ensemble_results = self.agg_ensemble(modelling_results=modelling_results_top, single_mode=True)

            top_ensemble_dict = self.__select_best_ensemble_method(ensemble_results)

            if len(top_ensemble_dict) == 0:
                self.logger.info(f'No improvement accomplished for {list(modelling_results_top.keys())} combination')
            else:
                current_ensemble_method = list(top_ensemble_dict)[0]
                best_ensemble_metric = top_ensemble_dict[current_ensemble_method]
                model_combination = list(modelling_results_top)[:top_K_models + 1]
                self.logger.info(
                    f'Accomplished metric improvement by {current_ensemble_method}:'
                    f'New best metric – {best_ensemble_metric}'
                )

                if self.best_ensemble_metric > 0:
                    self.experiment_results.update({'Ensemble_models': model_combination})
                    self.experiment_results.update({'Ensemble_method': current_ensemble_method})
                    self.experiment_results.update({'Best_ensemble_metric': best_ensemble_metric})
        return self.experiment_results

    def __select_best_ensemble_method(self, ensemble_results: dict):
        """
        A method that iteratively searches for an ensemble algorithm that improves the current best result

        Returns:
            sorted dictionary with structure {'Ensemble_models': 'best ensemble metric'}

        """
        top_ensemble_dict = {}
        for ensemble_method in ensemble_results:
            ensemble_dict = ensemble_results[ensemble_method]
            # ensemble_metrics = self.IndustrialModel.get_metrics(target=ensemble_dict['target'],
            #                                                     prediction_label=ensemble_dict['label'],
            #                                                     prediction_proba=ensemble_dict['proba'])

            ensemble_metrics = self.performance_analyzer.calculate_metrics(target=ensemble_dict['target'],
                                                                           predicted_labels=ensemble_dict['label'],
                                                                           predicted_probs=ensemble_dict['proba'],
                                                                           target_metrics=[self.metric])

            ensemble_metric = ensemble_metrics[self.metric]
            if ensemble_metric > self.best_base_results['Base_metric'] and ensemble_metric > self.best_ensemble_metric:
                self.best_ensemble_metric = ensemble_metric
                top_ensemble_dict.update({ensemble_method: ensemble_metric})
        return dict(sorted(top_ensemble_dict.items(), key=lambda x: x[1], reverse=True))

    def agg_ensemble(self, modelling_results: dict = None, single_mode: bool = False) -> dict:
        ensemble_dict = {}
        if single_mode:
            for strategy in self.ensemble_strategy:
                ensemble_dict.update({f'{strategy}': self._ensemble_by_method(modelling_results,
                                                                              strategy=strategy)})
        else:
            for generator in modelling_results:
                ensemble_dict[generator] = {}
                self.generator = generator
                for launch in modelling_results[generator]:
                    ensemble_dict[generator].update({launch: modelling_results[generator][launch]['metrics']})

                for strategy in self.ensemble_strategy:
                    ensemble_dict[generator].update({strategy: self._ensemble_by_method(modelling_results[generator],
                                                                                        strategy=strategy)})
        return ensemble_dict

    def _ensemble_by_method(self, predictions, strategy):
        transformed_predictions = self._check_predictions(predictions, strategy_name=strategy)
        average_proba_predictions = self.ensemble_strategy_dict[strategy](transformed_predictions, axis=1)

        if average_proba_predictions.shape[1] == 1:
            average_proba_predictions = np.concatenate([average_proba_predictions, 1 - average_proba_predictions],
                                                       axis=1)

        label_predictions = np.argmax(average_proba_predictions, axis=1)
        try:
            target = self.test_target
            metrics = self.performance_analyzer.calculate_metrics(target=target,
                                                                  predicted_labels=label_predictions,
                                                                  predicted_probs=average_proba_predictions,)
        except KeyError:
            metrics = None
            target = None

        return {'target': target,
                'label': label_predictions,
                'proba': average_proba_predictions,
                'metrics': metrics}

    def _check_predictions(self, predictions, strategy_name):
        """Check if the predictions array has the correct size.

        Args:
            predictions: array of shape (n_samples, n_classifiers). The votes obtained by each classifier
            for each sample.
            strategy_name: str. The name of the strategy used to ensemble the predictions.

        Returns:
            predictions: array of shape (n_samples, n_classifiers). The votes obtained by each classifier
            for each sample.

        Raises:
            ValueError: if the array do not contain exactly 3 dimensions: [n_samples, n_classifiers, n_classes]

        """
        if strategy_name in self.strategy_exclude_list:
            return predictions

        if type(predictions) == dict:
            try:
                list_proba = []
                for model_preds in predictions:
                    proba_frame = predictions[model_preds]
                    if isinstance(proba_frame, np.ndarray):
                        list_proba.append(proba_frame)
                    else:
                        try:
                            list_proba.append(proba_frame.values)
                        except KeyError:
                            self.target = proba_frame['Target'].values
                            if 'Preds' in proba_frame.columns:
                                filter_col = ['Target', 'Preds']
                            else:
                                filter_col = ['Target', 'Predicted_labels']
                            proba_frame = proba_frame.loc[:, ~proba_frame.columns.isin(filter_col)]
                            list_proba.append(proba_frame.values)
                return np.array(list_proba).transpose((1, 0, 2))
            except Exception as error:
                self.logger.error(f'Error in ensemble predictions: {error}')
                raise ValueError(
                    'predictions must contain 3 dimensions: ')



import hashlib
import os
import timeit

import numpy as np
import pandas as pd
import logging

from fedot_ind.core.architecture.utils.utils import PROJECT_PATH


class DataCacher:
    """Class responsible for caching data of ``pd.DataFrame`` type in pickle format.
    Args:
        data_type_prefix: a string prefix related to the data to be cached. For example, if data is related to
        modelling results, then the prefix can be 'ModellingResults'. Default prefix is 'Data'.
        cache_folder: path to the folder where data is going to be cached.
    Examples:
        >>> your_data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
        >>> data_cacher = DataCacher(data_type_prefix='data', cache_folder='your_path')
        >>> hashed_info = data_cacher.hash_info(dict(name='data', data=your_data))
        >>> data_cacher.cache_data(hashed_info, your_data)
        >>> data_cacher.load_data_from_cache(hashed_info)
    """

    def __init__(self, data_type_prefix: str = 'Data', cache_folder: str = None):
        self.data_type = data_type_prefix
        self.cache_folder = cache_folder
        self.cache_folder = self._init_cache_folder(cache_folder)

        self.logger = logging.getLogger('DataCacher')

    def _init_cache_folder(self, cache_folder):
        if cache_folder is None:
            cache_folder = os.path.join(PROJECT_PATH, 'cache')
        else:
            cache_folder = os.path.abspath(cache_folder)

        os.makedirs(cache_folder, exist_ok=True)
        return cache_folder

    def hash_info(self, data, **kwargs) -> str:
        """Method responsible for hashing distinct information about the data that is going to be cached.
        It utilizes md5 hashing algorithm.
        Args:
            kwargs: a set of keyword arguments to be used as distinct info about data.
        Returns:
            Hashed string.
        """
        key = ''.join([repr(arg) for arg in kwargs.values()]).encode('utf8')
        key += data.__str__().encode('utf8')
        hsh = hashlib.md5(key).hexdigest()[:10]
        return hsh

    def load_data_from_cache(self, hashed_info: str):
        """Method responsible for loading cached data.
        Args:
            hashed_info: hashed string of needed info about the data.
        """
        self.logger.info('Trying to load features from cache')

        start = timeit.default_timer()
        file_path = os.path.join(self.cache_folder, hashed_info + '.npy')
        data = np.load(file_path)
        elapsed_time = round(timeit.default_timer() - start, 5)
        print(f'{self.data_type} of {type(data)} type is loaded from cache in {elapsed_time} sec')
        return data

    def cache_data(self, hashed_info: str, data: pd.DataFrame):
        """Method responsible for saving cached data. It utilizes pickle format for saving data.
        Args:
            hashed_info: hashed string.
            data: pd.DataFrame.
        """
        self.logger.info('Caching features')
        cache_file = os.path.join(self.cache_folder, hashed_info)

        try:
            np.save(cache_file, data)

        except Exception as ex:
            print(f'Data was not cached due to error { ex }')


from fedot_ind.core.models.detection.utils.get_time import time_now


class AbstractDataOperation:

    def set_settings(self):
        self.name = ""
        self.print_logs: bool = True
        self._print_logs("Settings was set.")

    def load_data(self, object) -> None:
        raise NotImplementedError()

    def run_operation(self) -> None:
        self._print_logs(f"Start {self.operation}...")
        self._print_logs(f"{self.operation} finished!")

    def return_new_data(self):
        raise NotImplementedError()

    def _print_logs(self, log_message: str) -> None:
        if self.print_logs:
            print(f"---[{time_now()}] {self.name}: {log_message}")





import logging
from enum import Enum
from typing import Dict, List

from sklearn import preprocessing

from fedot_ind.core.metrics.metrics_implementation import *


class Metrics(Enum):
    roc_auc = ROCAUC
    f1 = F1
    precision = Precision
    accuracy = Accuracy
    logloss = Logloss
    rmse = RMSE,
    r2 = R2
    mae = MAE
    mse = MSE
    mape = MAPE


DEF_METRIC_LIST = ['f1', 'roc_auc', 'accuracy', 'logloss', 'precision']


class PerformanceAnalyzer:
    """Class responsible for calculating metrics for predictions.

    """
    def __init__(self):
        self.logger = logging.getLogger('PerformanceAnalyzer')

    @staticmethod
    def problem_and_metric_for_dataset(task_type: str) -> Union[List, None]:
        if task_type == 'classification':
            return ['f1', 'roc_auc', 'accuracy', 'logloss', 'precision']
        elif task_type == 'regression':
            return ['rmse', 'r2']
        else:
            return None

    @staticmethod
    def check_target(target, predictions):
        if type(target) is not np.ndarray:
            target = np.array(target)
        if type(target[0]) is str:
            enc = preprocessing.LabelEncoder().fit(target)
            converted_target = np.array(enc.transform(target))
            converted_predictions = np.array(enc.transform(predictions))
        else:
            converted_target = target
            converted_predictions = predictions

        if converted_target.dtype == object:
            converted_target = [int(x) for x in converted_target]

        return converted_target, converted_predictions

    def calculate_metrics(self,
                          target: Union[np.ndarray, List],
                          predicted_labels: Union[np.ndarray, list] = None,
                          predicted_probs: np.ndarray = None,
                          target_metrics: list = None) -> Dict:
        self.logger.info(f'Calculating metrics: {target_metrics}')
        try:
            if type(target[0]) is not float:
                labels_diff = max(target) - max(predicted_labels)

                if min(predicted_labels) != min(target):
                    if min(target) == -1:
                        np.place(predicted_labels, predicted_labels == 1, [-1])
                        np.place(predicted_labels, predicted_labels == 0, [1])

                if labels_diff > 0:
                    predicted_labels = predicted_labels + abs(labels_diff)
                else:
                    target = target + abs(labels_diff)
        except Exception:
            target = target

        if target_metrics is not None:
            metric_list = target_metrics
        else:
            metric_list = DEF_METRIC_LIST

        result_metric = []
        for metric_name in metric_list:
            chosen_metric = Metrics[metric_name].value
            try:
                score = chosen_metric(target=target,
                                      predicted_labels=predicted_labels,
                                      predicted_probs=predicted_probs).metric()
                score = round(score, 3)
                result_metric.append(score)
            except Exception as err:
                self.logger.info(f'Score cannot be calculated for {metric_name} metric')
                self.logger.info(err)
                result_metric.append(0)

        result_dict = dict(zip(metric_list, result_metric))
        self.logger.info(f'Metrics are: {result_dict}')

        return result_dict



import logging
import warnings

import numpy as np
import pandas as pd
from fedot.api.main import Fedot

warnings.simplefilter(action='ignore', category=FutureWarning)


class Booster:
    """Class to implement error correction model (ECM) for classification prediction improvement.

    Args:
        features_train: X_train
        target_train: y_train
        base_predict: prediction, derived from main model (Quantile, Spectral, Topological, or Wavelet)
        timeout: defines the amount of time to compose and tune regression model
        threshold: parameter used as round boundary for custom_round() method
        n_cycles: number of boosting cycles
        reshape_flag: ...

    Examples:
        >>> X_train = pd.read_csv('X_train.csv')
        >>> y_train = pd.read_csv('y_train.csv')
        >>> base_predict = pd.read_csv('base_predict.csv')

        >>> y_train = y_train.values
        >>> base_predict = base_predict.values

        >>> booster = Booster(features_train=X_train,
                              target_train=y_train,
                              base_predict=base_predict,
                              timeout=1)

        >>> result = booster.fit()
    """

    def __init__(self, features_train: np.ndarray,
                 target_train: np.array,
                 base_predict: np.array,
                 timeout: int,
                 threshold: float = 0.5,
                 reshape_flag: bool = False,
                 n_cycles: int = 3):
        self.logger = logging.getLogger(self.__class__.__name__)

        self.X_train = features_train
        self.target_train = target_train
        self.base_predict = base_predict
        self.threshold = threshold
        self.timeout = max(3, round(timeout / 4))
        # self.timeout = 1
        self.reshape_flag = reshape_flag
        self.n_cycles = n_cycles
        if self.reshape_flag:
            self.target_train = self.target_train.reshape(-1)

        self.ecm_targets = [self.target_train, ]
        self.ecm_predictions = [self.base_predict, ]

    def fit(self) -> tuple:
        """Method to fit error correction models.

        Returns:
            tuple: tuple with ensemble prediction and ensemble model
        """
        self.logger.info('Started boosting')
        model_list = list()

        for i in range(self.n_cycles):
            self.logger.info(f'Starting cycle {1 + i} of boosting')
            target_diff = np.subtract(previous_predict=self.ecm_predictions[i],
                                      previous_target=self.ecm_targets[i])
            self.ecm_targets.append(target_diff)

            prediction, fedot_model = self.api_model(target_diff=target_diff)
            self.ecm_predictions.append(prediction)
            model_list.append(fedot_model)

        final_prediction, model_ensemble = self.ensemble(features=self.ecm_predictions[1:])

        return final_prediction, model_list, model_ensemble

    def api_model(self, target_diff: np.ndarray) -> tuple:
        """Method used to initiate FEDOT AutoML model to solve regression problem for boosting stage

        Args:
            target_diff: target difference between previous prediction and target

        Returns:
            Tuple with prediction and FEDOT model
        """
        fedot_model = Fedot(problem='regression', seed=42,
                            timeout=self.timeout, max_depth=10,
                            max_arity=4,
                            cv_folds=2,
                            logging_level=20, n_jobs=2)

        fedot_model.fit(self.X_train, target_diff)
        prediction = fedot_model.predict(self.X_train)

        if self.reshape_flag:
            prediction = prediction.reshape(-1)

        return prediction, fedot_model

    def ensemble(self, features: list) -> tuple:
        """Method that ensembles results of all stages of boosting. Depending on number of classes ensemble method
        could be a genetic AutoML model by FEDOT (for binary problem) or SUM method (for multi-class problem).

        Args:
            features: list of predictions from all boosting stages

        Returns:
            Tuple with ensemble prediction and ensemble model

        """
        self.logger.info('Starting to ensemble boosting results')

        ensemble_model = Fedot(problem='regression',
                               timeout=self.timeout,
                               seed=20,
                               logging_level=20,
                               n_jobs=6)
        if self.reshape_flag:
            features = np.hstack(features)
            target = self.target_train

            ensemble_model.fit(features, target)
            ensemble_prediction = ensemble_model.predict(features)
            ensemble_prediction = ensemble_prediction.reshape(-1)
            return ensemble_prediction, ensemble_model

        else:
            ensemble_prediction = sum(features)
            return ensemble_prediction, None

    @staticmethod
    def proba_to_vector(matrix: np.array):
        """Method to convert probability matrix to vector of labels.

        Args:
            matrix: probability matrix

        Returns:
            Probability matrix received as a result of prediction_proba for multi-class problem

        """
        vector = np.array([x.argmax() + x[x.argmax()] for x in matrix])
        return vector

    def custom_round(self, num: float) -> int:
        """Custom round method with predefined threshold.

        Args:
            num: number to round

        Returns:
            Rounded number
        """
        thr = self.threshold
        if num - int(num) >= thr:
            return int(num) + 1
        return int(num)

